"""Backup related CLI commands."""

import logging
from pathlib import Path
from typing import Iterable

import rich_click as click
import housekeeper.store.models as hk_models

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.apps.slurm.slurm_api import SlurmAPI
from cg.cli.utils import CLICK_CONTEXT_SETTINGS
from cg.constants.cli_options import DRY_RUN
from cg.constants.constants import SequencingRunDataAvailability
from cg.constants.housekeeper_tags import SequencingFileTag
from cg.exc import (
    DsmcAlreadyRunningError,
    FlowCellError,
    IlluminaRunAlreadyBackedUpError,
    IlluminaRunEncryptionError,
    PdcError,
)
from cg.meta.backup.backup import SpringBackupAPI
from cg.meta.encryption.encryption import EncryptionAPI, SpringEncryptionAPI
from cg.meta.tar.tar import TarAPI
from cg.models.cg_config import CGConfig
from cg.models.run_devices.illumina_run_directory_data import (
    IlluminaRunDirectoryData,
    get_sequencing_runs_from_path,
)
from cg.services.illumina.backup.backup_service import IlluminaBackupService
from cg.services.illumina.backup.encrypt_service import (
    IlluminaRunEncryptionService,
)
from cg.services.pdc_service.pdc_service import PdcService
from cg.store.exc import EntryNotFoundError
from cg.store.models import IlluminaSequencingRun, Sample
from cg.store.store import Store

LOG = logging.getLogger(__name__)


@click.group(context_settings=CLICK_CONTEXT_SETTINGS)
@click.pass_obj
def backup(context: CGConfig):
    """Backup utilities"""
    pass


@backup.command("illumina-runs")
@DRY_RUN
@click.pass_obj
def backup_illumina_runs(context: CGConfig, dry_run: bool):
    """Back-up Illumina runs."""
    pdc_service = context.pdc_service
    pdc_service.dry_run = dry_run
    encryption_api = EncryptionAPI(binary_path=context.encryption.binary_path, dry_run=dry_run)
    tar_api = TarAPI(binary_path=context.tar.binary_path, dry_run=dry_run)
    backup_service = IlluminaBackupService(
        encryption_api=encryption_api,
        pdc_archiving_directory=context.illumina_backup_service.pdc_archiving_directory,
        status_db=context.status_db,
        tar_api=tar_api,
        pdc_service=pdc_service,
        sequencing_runs_dir=context.run_instruments.illumina.sequencing_runs_dir,
        dry_run=dry_run,
    )
    backup_service.dry_run = dry_run
    status_db: Store = context.status_db
    runs_dir_data: list[IlluminaRunDirectoryData] = get_sequencing_runs_from_path(
        sequencing_run_dir=Path(context.run_instruments.illumina.sequencing_runs_dir)
    )
    for run_dir_data in runs_dir_data:
        try:
            sequencing_run: IlluminaSequencingRun = (
                status_db.get_illumina_sequencing_run_by_device_internal_id(run_dir_data.id)
            )
            backup_service.start_run_backup(
                run_dir_data=run_dir_data,
                sequencing_run=sequencing_run,
                status_db=status_db,
                binary_path=context.encryption.binary_path,
                encryption_dir=Path(context.encryption.encryption_dir),
                pigz_binary_path=context.pigz.binary_path,
                sbatch_parameter=context.illumina_backup_service.slurm_flow_cell_encryption.dict(),
            )
        except EntryNotFoundError as error:
            logging.error(f"{error}")
            continue
        except (
            DsmcAlreadyRunningError,
            IlluminaRunAlreadyBackedUpError,
            IlluminaRunEncryptionError,
            PdcError,
        ) as error:
            logging.error(f"{error}")


@backup.command("encrypt-illumina-runs")
@DRY_RUN
@click.pass_obj
def encrypt_illumina_runs(context: CGConfig, dry_run: bool):
    """Encrypt illumina runs."""
    status_db: Store = context.status_db
    runs: list[IlluminaRunDirectoryData] = get_sequencing_runs_from_path(
        sequencing_run_dir=Path(context.run_instruments.illumina.sequencing_runs_dir)
    )
    for run in runs:
        try:
            sequencing_run: IlluminaSequencingRun = (
                status_db.get_illumina_sequencing_run_by_device_internal_id(run.id)
            )
        except EntryNotFoundError as error:
            LOG.error(f"{error}")
            continue
        if sequencing_run.has_backup:
            LOG.debug(f"Run: {run.id} is already backed-up")
            continue
        illumina_run_encryption_service = IlluminaRunEncryptionService(
            binary_path=context.encryption.binary_path,
            dry_run=dry_run,
            encryption_dir=Path(context.encryption.encryption_dir),
            run_dir_data=run,
            pigz_binary_path=context.pigz.binary_path,
            slurm_api=SlurmAPI(),
            sbatch_parameter=context.illumina_backup_service.slurm_flow_cell_encryption.dict(),
            tar_api=TarAPI(binary_path=context.tar.binary_path, dry_run=dry_run),
        )
        try:
            illumina_run_encryption_service.start_encryption()
        except (FlowCellError, IlluminaRunEncryptionError) as error:
            logging.error(f"{error}")


@backup.command("fetch-illumina-run")
@click.option("-f", "--flow-cell-id", help="Retrieve a specific flow cell, ex. 'HCK2KDSXX'")
@DRY_RUN
@click.pass_obj
def fetch_illumina_run(context: CGConfig, dry_run: bool, flow_cell_id: str | None = None):
    """Fetch the first Illumina run in the requested queue from backup."""

    pdc_service = context.pdc_service
    pdc_service.dry_run = dry_run
    encryption_api = EncryptionAPI(binary_path=context.encryption.binary_path, dry_run=dry_run)
    tar_api = TarAPI(binary_path=context.tar.binary_path, dry_run=dry_run)
    context.meta_apis["backup_api"] = IlluminaBackupService(
        encryption_api=encryption_api,
        pdc_archiving_directory=context.illumina_backup_service.pdc_archiving_directory,
        status_db=context.status_db,
        tar_api=tar_api,
        pdc_service=pdc_service,
        sequencing_runs_dir=context.run_instruments.illumina.sequencing_runs_dir,
        dry_run=dry_run,
    )
    backup_api: IlluminaBackupService = context.meta_apis["backup_api"]

    status_db: Store = context.status_db
    sequencing_run: IlluminaSequencingRun | None = None
    if not flow_cell_id:
        LOG.info("Fetching first sequencing run in queue")
    try:
        if flow_cell_id:
            sequencing_run: IlluminaSequencingRun = (
                status_db.get_illumina_sequencing_run_by_device_internal_id(flow_cell_id)
            )
    except EntryNotFoundError as error:
        LOG.error(f"{error}")
        raise click.Abort

    retrieval_time: float | None = backup_api.fetch_sequencing_run(sequencing_run)

    if retrieval_time:
        hours = retrieval_time / 60 / 60
        LOG.info(f"Retrieval time: {hours:.1}h")
        return

    if not dry_run and sequencing_run:
        LOG.info(
            f"{sequencing_run}: updating sequencing run data availability to {SequencingRunDataAvailability.REQUESTED}"
        )
        status_db.update_illumina_sequencing_run_data_availability(
            sequencing_run=sequencing_run, data_availability=SequencingRunDataAvailability.REQUESTED
        )


@backup.command("archive-spring-files")
@DRY_RUN
@click.pass_context
@click.pass_obj
def archive_spring_files(config: CGConfig, context: click.Context, dry_run: bool):
    """Archive spring files to PDC."""
    housekeeper_api: HousekeeperAPI = config.housekeeper_api
    LOG.info("Getting all spring files from Housekeeper.")
    spring_files: Iterable[hk_models.File] = housekeeper_api.files(
        tags=[SequencingFileTag.SPRING]
    ).filter(hk_models.File.path.contains(f"{config.environment}/{config.demultiplex.out_dir}"))
    for spring_file in spring_files:
        LOG.info(f"Attempting encryption and PDC archiving for file {spring_file.path}")
        if Path(spring_file.path).exists():
            context.invoke(archive_spring_file, spring_file_path=spring_file.path, dry_run=dry_run)
        else:
            LOG.warning(
                f"Spring file {spring_file.path} found in Housekeeper, but not on disk! Archiving process skipped!"
            )


@backup.command("archive-spring-file")
@click.argument("spring-file-path", type=click.Path(exists=True))
@DRY_RUN
@click.pass_obj
def archive_spring_file(config: CGConfig, spring_file_path: str, dry_run: bool):
    """Archive a spring file to PDC."""
    housekeeper_api: HousekeeperAPI = config.housekeeper_api
    pdc_service: PdcService = PdcService(binary_path=config.pdc.binary_path, dry_run=dry_run)
    encryption_api: SpringEncryptionAPI = SpringEncryptionAPI(
        binary_path=config.encryption.binary_path,
        dry_run=dry_run,
    )
    spring_backup_api: SpringBackupAPI = SpringBackupAPI(
        encryption_api=encryption_api,
        hk_api=housekeeper_api,
        pdc_service=pdc_service,
        dry_run=dry_run,
    )
    LOG.debug("Start spring encryption/backup")
    spring_backup_api.encrypt_and_archive_spring_file(Path(spring_file_path))


@backup.command("retrieve-spring-files")
@DRY_RUN
@click.option("-s", "--sample-id", "object_type", flag_value="sample", type=str)
@click.option("-c", "--case-id", "object_type", flag_value="case", type=str)
@click.option("-f", "--flow-cell-id", "object_type", flag_value="flow_cell", type=str)
@click.argument("identifier", type=str)
@click.pass_context
@click.pass_obj
def retrieve_spring_files(
    config: CGConfig,
    context: click.Context,
    object_type: str,
    identifier: str,
    dry_run: bool,
):
    """Retrieve all spring files for a given identity."""
    status_api: Store = config.status_db
    housekeeper_api: HousekeeperAPI = config.housekeeper_api

    samples: list[Sample] = status_api.get_samples_by_identifier(
        object_type=object_type, identifier=identifier
    )

    for sample in samples:
        latest_version: hk_models.Version = housekeeper_api.last_version(bundle=sample.internal_id)
        spring_files: Iterable[hk_models.File] = housekeeper_api.files(
            bundle=sample.internal_id,
            tags=[SequencingFileTag.SPRING],
            version=latest_version.id,
        )
        for spring_file in spring_files:
            context.invoke(retrieve_spring_file, spring_file_path=spring_file.path, dry_run=dry_run)


@backup.command("retrieve-spring-file")
@click.argument("spring-file-path", type=click.Path())
@DRY_RUN
@click.pass_obj
def retrieve_spring_file(config: CGConfig, spring_file_path: str, dry_run: bool):
    """Retrieve a spring file from PDC."""
    LOG.info(f"Attempting PDC retrieval and decryption file {spring_file_path}")
    housekeeper_api: HousekeeperAPI = config.housekeeper_api
    pdc_service: PdcService = PdcService(binary_path=config.pdc.binary_path, dry_run=dry_run)
    encryption_api: SpringEncryptionAPI = SpringEncryptionAPI(
        binary_path=config.encryption.binary_path,
        dry_run=dry_run,
    )
    LOG.debug(f"Start spring retrieval if not dry run mode={dry_run}")
    spring_backup_api: SpringBackupAPI = SpringBackupAPI(
        encryption_api=encryption_api,
        hk_api=housekeeper_api,
        pdc_service=pdc_service,
        dry_run=dry_run,
    )
    spring_backup_api.retrieve_and_decrypt_spring_file(Path(spring_file_path))



from abc import abstractmethod, ABC
from pathlib import Path


class PathNameFormatter(ABC):
    """
    Abstract class that encapsulates the logic required for formatting the path name.
    """

    @abstractmethod
    def format_file_path(self, file_path: Path, provided_id: str, provided_name: str) -> Path:
        """Format the file path."""
        pass



from pathlib import Path
from typing import Callable

from pydantic import BaseModel


class FastqFileMeta(BaseModel):
    path: Path | None = None
    lane: int
    read_direction: int
    undetermined: bool | None = None
    flow_cell_id: str


def _get_header_meta_casava_five_parts(parts: list[str]) -> FastqFileMeta:
    return FastqFileMeta(
        lane=parts[1], flow_cell_id="XXXXXX", read_direction=parts[-1].split("/")[-1]
    )


def _get_header_meta_casava_ten_parts(parts: list[str]) -> FastqFileMeta:
    return FastqFileMeta(
        lane=parts[3], flow_cell_id=parts[2], read_direction=parts[6].split(" ")[-1]
    )


def _get_header_meta_casava_seven_parts(parts: list[str]) -> FastqFileMeta:
    return FastqFileMeta(
        lane=parts[3], flow_cell_id=parts[2], read_direction=parts[-1].split("/")[-1]
    )


class GetFastqFileMeta:
    header_format: dict[int, Callable] = {
        5: _get_header_meta_casava_five_parts,
        7: _get_header_meta_casava_seven_parts,
        10: _get_header_meta_casava_ten_parts,
    }



"""Module for the Illumina encrypt run service test."""

import logging

import shutil
from pathlib import Path


import pytest

from cg.exc import IlluminaRunEncryptionError, FlowCellError
from cg.services.illumina.backup.encrypt_service import (
    IlluminaRunEncryptionService,
)
from cg.models.run_devices.illumina_run_directory_data import IlluminaRunDirectoryData


def test_illumina_run_encryption_service(
    illumina_run_encryption_service: IlluminaRunEncryptionService,
):
    """Tests instantiating an IlluminaRunEncryptionService."""
    # GIVEN an IlluminaRunEncryptionService

    # WHEN instantiating the service

    # THEN return a IlluminaRunEncryptionService object
    assert isinstance(illumina_run_encryption_service, IlluminaRunEncryptionService)


def test_get_run_symmetric_encryption_command(
    illumina_run_encryption_service: IlluminaRunEncryptionService,
    output_file_path: Path,
    temporary_passphrase: Path,
):
    # GIVEN a IlluminaRunEncryptionService

    # WHEN getting the command
    command: str = illumina_run_encryption_service.get_run_symmetric_encryption_command(
        output_file=output_file_path, passphrase_file_path=temporary_passphrase
    )

    # THEN return a string
    assert isinstance(command, str)


def test_get_run_symmetric_decryption_command(
    illumina_run_encryption_service: IlluminaRunEncryptionService,
    input_file_path: Path,
    temporary_passphrase: Path,
):
    # GIVEN a IlluminaRunEncryptionService

    # WHEN getting the command
    command: str = illumina_run_encryption_service.get_run_symmetric_decryption_command(
        input_file=input_file_path, passphrase_file_path=temporary_passphrase
    )

    # THEN return a string
    assert isinstance(command, str)


def test_create_pending_file(
    illumina_run_encryption_service: IlluminaRunEncryptionService,
):
    # GIVEN a IlluminaRunEncryptionService

    # WHEN checking if encryption is possible
    illumina_run_encryption_service.run_encryption_dir.mkdir(parents=True)
    illumina_run_encryption_service.dry_run = False
    illumina_run_encryption_service.create_pending_file(
        pending_path=illumina_run_encryption_service.pending_file_path
    )

    # THEN pending file should exist
    assert illumina_run_encryption_service.pending_file_path.exists()

    # Clean-up
    shutil.rmtree(illumina_run_encryption_service.run_encryption_dir)


def test_create_pending_file_when_dry_run(
    illumina_run_encryption_service: IlluminaRunEncryptionService,
):
    # GIVEN a IlluminaRunEncryptionService

    # WHEN checking if encryption is possible
    illumina_run_encryption_service.create_pending_file(
        pending_path=illumina_run_encryption_service.pending_file_path
    )

    # THEN pending file should not exist
    assert not illumina_run_encryption_service.pending_file_path.exists()


def test_is_encryption_possible(
    illumina_run_encryption_service: IlluminaRunEncryptionService, mocker
):
    # GIVEN a IlluminaRunEncryptionService

    # GIVEN that sequencing is ready
    mocker.patch.object(IlluminaRunDirectoryData, "is_sequencing_run_ready")
    IlluminaRunDirectoryData.is_sequencing_run_ready.return_value = True

    # WHEN checking if encryption is possible
    is_possible: bool = illumina_run_encryption_service.is_encryption_possible()

    # THEN return True
    assert is_possible


def test_is_encryption_possible_when_sequencing_not_ready(
    caplog,
    illumina_run_encryption_service: IlluminaRunEncryptionService,
    flow_cell_name: str,
    mocker,
):
    caplog.set_level(logging.ERROR)

    # GIVEN a IlluminaRunEncryptionService

    # GIVEN that sequencing is not ready
    mocker.patch.object(IlluminaRunDirectoryData, "is_sequencing_run_ready")
    IlluminaRunDirectoryData.is_sequencing_run_ready.return_value = False

    # WHEN checking if encryption is possible
    with pytest.raises(FlowCellError):
        illumina_run_encryption_service.is_encryption_possible()

        # THEN error should be raised


def test_is_encryption_possible_when_encryption_is_completed(
    caplog,
    illumina_run_encryption_service: IlluminaRunEncryptionService,
    mocker,
):
    # GIVEN a IlluminaRunEncryptionService

    # GIVEN that sequencing is ready
    mocker.patch.object(IlluminaRunDirectoryData, "is_sequencing_run_ready")
    IlluminaRunDirectoryData.is_sequencing_run_ready.return_value = True

    # GIVEN that encryption is completed
    illumina_run_encryption_service.run_encryption_dir.mkdir(parents=True)
    illumina_run_encryption_service.complete_file_path.touch()

    # WHEN checking if encryption is possible
    with pytest.raises(IlluminaRunEncryptionError):
        illumina_run_encryption_service.is_encryption_possible()

        # THEN error should be raised

    # Clean-up
    shutil.rmtree(illumina_run_encryption_service.run_encryption_dir)


def test_is_encryption_possible_when_encryption_is_pending(
    caplog,
    illumina_run_encryption_service: IlluminaRunEncryptionService,
    mocker,
):
    # GIVEN a IlluminaRunEncryptionService

    # GIVEN that sequencing is ready
    mocker.patch.object(IlluminaRunDirectoryData, "is_sequencing_run_ready")
    IlluminaRunDirectoryData.is_sequencing_run_ready.return_value = True

    # GIVEN that encryption is pending
    illumina_run_encryption_service.run_encryption_dir.mkdir(parents=True)
    illumina_run_encryption_service.pending_file_path.touch()

    # WHEN checking if encryption is possible
    with pytest.raises(IlluminaRunEncryptionError):
        illumina_run_encryption_service.is_encryption_possible()

        # THEN an error should be raised

    # Clean-up
    shutil.rmtree(illumina_run_encryption_service.run_encryption_dir)


def test_encrypt_run(
    caplog,
    illumina_run_encryption_service: IlluminaRunEncryptionService,
    mocker,
    sbatch_job_number: int,
):
    caplog.set_level(logging.INFO)
    # GIVEN a IlluminaRunEncryptionService

    # GIVEN that sequencing is ready
    mocker.patch.object(IlluminaRunDirectoryData, "is_sequencing_run_ready")
    IlluminaRunDirectoryData.is_sequencing_run_ready.return_value = True

    # WHEN encrypting an Illumina run
    illumina_run_encryption_service.encrypt_run()

    # THEN sbatch should be submitted
    assert f"Run encryption running as job {sbatch_job_number}" in caplog.text


def test_start_encryption(
    caplog,
    illumina_run_encryption_service: IlluminaRunEncryptionService,
    mocker,
    sbatch_job_number: int,
):
    caplog.set_level(logging.INFO)
    # GIVEN a IlluminaRunEncryptionService

    # GIVEN that sequencing is ready
    mocker.patch.object(IlluminaRunDirectoryData, "is_sequencing_run_ready")
    IlluminaRunDirectoryData.is_sequencing_run_ready.return_value = True

    # WHEN trying to start encrypting an Illumina run
    illumina_run_encryption_service.start_encryption()

    # THEN sbatch should be submitted
    assert f"Run encryption running as job {sbatch_job_number}" in caplog.text



from pathlib import Path

from cg.constants.constants import FileFormat
from cg.io.controller import ReadFile
from cg.meta.workflow.balsamic import BalsamicAnalysisAPI


def get_balsamic_raw_data() -> tuple[dict, dict]:
    """Extracts mock BALSAMIC analysis data."""

    config: dict = ReadFile.get_content_from_file(
        file_format=FileFormat.YAML,
        file_path=Path("tests", "fixtures", "apps", "balsamic", "tga_case", "config.json"),
    )
    metrics: dict = ReadFile.get_content_from_file(
        file_format=FileFormat.YAML,
        file_path=Path(
            "tests", "fixtures", "apps", "balsamic", "tga_case", "metrics_deliverables.yaml"
        ),
    )
    return config, metrics


class MockBalsamicAnalysis(BalsamicAnalysisAPI):
    """Mock a BALSAMIC analysis object."""

    def get_latest_metadata(self, case_id: str):
        """Return the latest metadata of a specific BALSAMIC case."""

        config, metrics = get_balsamic_raw_data()
        return self.parse_analysis(config, metrics)

    def get_bundle_deliverables_type(self, case_id: str) -> str:
        """Return the analysis type for a case."""

        return "tumor_normal_panel"



"""Test module cg.cli.upload.nipt.statina"""

import logging

from click.testing import CliRunner

from cg.cli.upload.nipt.statina import batch
from cg.meta.upload.nipt import NiptUploadAPI
from cg.models.cg_config import CGConfig


class MockStatinaUploadFiles:
    def json(self, *args, **kwargs):
        return ""


def test_nipt_statina_upload_case(upload_context: CGConfig, cli_runner: CliRunner, caplog, mocker):
    """Tests CLI command to upload a single case"""

    caplog.set_level(logging.DEBUG)
    # GIVEN a specified NIPT case
    case_id = "angrybird"

    # WHEN adding a result file of a specified NIPT case
    mocker.patch.object(NiptUploadAPI, "get_statina_files", return_value=MockStatinaUploadFiles())
    mocker.patch.object(NiptUploadAPI, "upload_to_statina_database")
    mocker.patch.object(NiptUploadAPI, "sequencing_run_passed_qc_value", return_value=True)
    result = cli_runner.invoke(batch, [case_id], obj=upload_context)

    # THEN the nipt upload should start and exit without errors
    assert "*** Statina UPLOAD START ***" in caplog.text
    assert result.exit_code == 0


def test_nipt_statina_upload_case_dry(
    upload_context: CGConfig, cli_runner: CliRunner, caplog, mocker
):
    """Tests CLI command to upload a single case"""

    caplog.set_level(logging.DEBUG)
    # GIVEN a specified NIPT case
    case_id = "angrybird"

    # WHEN adding a result file of a specified NIPT case (dry run)
    mocker.patch.object(NiptUploadAPI, "get_statina_files", return_value=MockStatinaUploadFiles())
    mocker.patch.object(NiptUploadAPI, "upload_to_statina_database")
    result = cli_runner.invoke(batch, ["--dry-run", case_id], obj=upload_context)

    # THEN the nipt upload should start and exit without errors
    assert "*** Statina UPLOAD START ***" in caplog.text
    assert result.exit_code == 0



import shutil
from pathlib import Path

from cg.cli.demultiplex.demux import create_manifest_files
from cg.constants.demultiplexing import DemultiplexingDirsAndFiles
from cg.io.csv import read_csv
from cg.utils.files import get_all_files_in_directory_tree


def test_create_manifest_files_true(
    cli_runner, tmp_flow_cells_directory_ready_for_demultiplexing: Path
):
    """Test that manifest files are created for flow cells where the sequencing is complete."""
    # GIVEN two flowcell directories
    first_flowcell_directory: Path = tmp_flow_cells_directory_ready_for_demultiplexing
    second_flowcell_directory = Path(
        shutil.copytree(
            first_flowcell_directory,
            Path(tmp_flow_cells_directory_ready_for_demultiplexing.parent, "flowcell_2"),
        )
    )

    # WHEN creating manifest files
    cli_runner.invoke(
        cli=create_manifest_files,
        args=["--source-directory", first_flowcell_directory.parent.as_posix()],
    )

    for flow_cell in [first_flowcell_directory, second_flowcell_directory]:
        manifest_file = Path(flow_cell, DemultiplexingDirsAndFiles.CG_FILE_MANIFEST)

        # THEN the manifest file should exist
        assert manifest_file.exists()

        # THEN the manifest file should contain all files in the flowcell directory
        assert_manifest_file_contains_all_files(
            manifest_file=manifest_file, directory_of_interest=flow_cell
        )


def test_create_manifest_files_false(
    cli_runner, tmp_flow_cells_directory_ready_for_demultiplexing: Path
):
    """Test that manifest files are not created for flow cells where the sequencing is not complete."""
    # GIVEN two flowcell directories with missing CopyComplete.txt
    first_flowcell_directory = tmp_flow_cells_directory_ready_for_demultiplexing
    Path(first_flowcell_directory, DemultiplexingDirsAndFiles.COPY_COMPLETE).unlink()
    second_flowcell_directory = Path(
        shutil.copytree(
            first_flowcell_directory,
            Path(tmp_flow_cells_directory_ready_for_demultiplexing.parent, "flowcell_2"),
        )
    )

    # WHEN creating manifest files
    cli_runner.invoke(
        cli=create_manifest_files,
        args=["--source-directory", first_flowcell_directory.parent.as_posix()],
    )

    for flow_cell in [first_flowcell_directory, second_flowcell_directory]:
        manifest_file = Path(flow_cell, DemultiplexingDirsAndFiles.CG_FILE_MANIFEST)

        # THEN the manifest file should not exist
        assert not manifest_file.exists()


def test_create_manifest_files_true_nanopore_data(
    cli_runner, nanopore_flow_cells_dir: Path, tmp_path
):
    """Test that manifest files are created for flow cells where the sequencing is complete."""
    # GIVEN a Nanopore run with two samples
    tmp_nanopore_directory = Path(
        shutil.copytree(
            nanopore_flow_cells_dir, Path(tmp_path, nanopore_flow_cells_dir.name).as_posix()
        )
    )
    sample_directories: list[Path] = list(Path(tmp_nanopore_directory).glob("*/*/*"))
    assert len(sample_directories) == 2

    # WHEN creating manifest files
    cli_runner.invoke(
        cli=create_manifest_files,
        args=["--source-directory", f"{tmp_nanopore_directory}/*/*"],
    )

    for sample_directory in sample_directories:
        manifest_file = Path(sample_directory, DemultiplexingDirsAndFiles.CG_FILE_MANIFEST)

        # THEN the manifest file should exist
        assert manifest_file.exists()

        # THEN the manifest file should contain all files in the flowcell directory
        assert_manifest_file_contains_all_files(
            manifest_file=manifest_file, directory_of_interest=sample_directory
        )


def assert_manifest_file_contains_all_files(
    manifest_file: Path, directory_of_interest: Path
) -> None:
    """Assert that the manifest file contains all files in the directory of interest."""
    all_files_in_directory: list[Path] = get_all_files_in_directory_tree(directory_of_interest)
    all_files_in_directory.remove(Path(manifest_file.name))
    files_in_manifest: list[Path] = [
        Path(file[0].strip()) for file in read_csv(delimiter="\t", file_path=manifest_file)
    ]
    assert all(file in files_in_manifest for file in all_files_in_directory)



from cg.services.orders.store_order_services.store_microbial_order import StoreMicrobialOrderService
from cg.store.models import Case
import datetime as dt
from cg.constants import DataDelivery
from cg.constants.constants import Workflow
from cg.models.orders.constants import OrderType
from cg.models.orders.order import OrderIn
from cg.models.orders.sample_base import ControlEnum
from cg.models.orders.samples import SarsCov2Sample
from cg.store.models import Customer, Sample
from cg.store.store import Store


def test_microbial_samples_to_status(
    microbial_order_to_submit: dict, store_microbial_order_service: StoreMicrobialOrderService
):
    # GIVEN microbial order with three samples
    order = OrderIn.parse_obj(microbial_order_to_submit, OrderType.MICROSALT)

    # WHEN parsing for status
    data = store_microbial_order_service.order_to_status(order=order)

    # THEN it should pick out samples and relevant information
    assert len(data["samples"]) == 5
    assert data["customer"] == "cust002"
    assert data["order"] == "Microbial samples"
    assert data["comment"] == "Order comment"
    assert data["data_analysis"] == Workflow.MICROSALT
    assert data["data_delivery"] == str(DataDelivery.FASTQ)

    # THEN first sample should contain all the relevant data from the microbial order
    sample_data = data["samples"][0]
    assert sample_data["priority"] == "research"
    assert sample_data["name"] == "all-fields"
    assert sample_data.get("internal_id") is None
    assert sample_data["organism_id"] == "M.upium"
    assert sample_data["reference_genome"] == "NC_111"
    assert sample_data["application"] == "MWRNXTR003"
    assert sample_data["comment"] == "plate comment"
    assert sample_data["volume"] == "1"


def test_sarscov2_samples_to_status(
    sarscov2_order_to_submit: dict, store_microbial_order_service: StoreMicrobialOrderService
):
    # GIVEN sarscov2 order with three samples
    order = OrderIn.parse_obj(sarscov2_order_to_submit, OrderType.SARS_COV_2)

    # WHEN parsing for status
    data = store_microbial_order_service.order_to_status(order=order)

    # THEN it should pick out samples and relevant information
    assert len(data["samples"]) == 6
    assert data["customer"] == "cust002"
    assert data["order"] == "Sars-CoV-2 samples"
    assert data["comment"] == "Order comment"
    assert data["data_analysis"] == Workflow.MUTANT
    assert data["data_delivery"] == str(DataDelivery.FASTQ)

    # THEN first sample should contain all the relevant data from the microbial order
    sample_data = data["samples"][0]
    assert sample_data.get("internal_id") is None
    assert sample_data["priority"] == "research"
    assert sample_data["application"] == "VWGDPTR001"
    assert sample_data["comment"] == "plate comment"
    assert sample_data["name"] == "all-fields"
    assert sample_data["organism_id"] == "SARS CoV-2"
    assert sample_data["reference_genome"] == "NC_111"
    assert sample_data["volume"] == "1"


def test_store_microbial_samples(
    base_store: Store,
    microbial_status_data: dict,
    ticket_id: str,
    store_microbial_order_service: StoreMicrobialOrderService,
):
    # GIVEN a basic store with no samples and a microbial order and one Organism
    assert not base_store._get_query(table=Sample).first()
    assert base_store._get_query(table=Case).count() == 0
    assert base_store.get_all_organisms().count() == 1

    # WHEN storing the order
    new_samples = store_microbial_order_service.store_items_in_status(
        customer_id=microbial_status_data["customer"],
        order=microbial_status_data["order"],
        ordered=dt.datetime.now(),
        ticket_id=ticket_id,
        items=microbial_status_data["samples"],
        comment="",
        data_analysis=Workflow.MICROSALT,
        data_delivery=DataDelivery.FASTQ_QC,
    )

    # THEN it should store the samples under a case (case) and the used previously unknown
    # organisms
    assert new_samples
    assert base_store._get_query(table=Case).count() == 1
    assert len(new_samples) == 5
    assert len(base_store._get_query(table=Sample).all()) == 5
    assert base_store.get_all_organisms().count() == 3


def test_store_microbial_case_data_analysis_stored(
    base_store: Store,
    microbial_status_data: dict,
    ticket_id: str,
    store_microbial_order_service: StoreMicrobialOrderService,
):
    # GIVEN a basic store with no samples and a microbial order and one Organism
    assert not base_store._get_query(table=Sample).first()
    assert base_store._get_query(table=Case).count() == 0

    # WHEN storing the order
    store_microbial_order_service.store_items_in_status(
        customer_id=microbial_status_data["customer"],
        order=microbial_status_data["order"],
        ordered=dt.datetime.now(),
        ticket_id=ticket_id,
        items=microbial_status_data["samples"],
        comment="",
        data_analysis=Workflow.MICROSALT,
        data_delivery=DataDelivery.FASTQ_QC,
    )

    # THEN store the samples under a case with the microbial data_analysis type on case level
    assert len(base_store._get_query(table=Sample).all()) > 0
    assert base_store._get_query(table=Case).count() == 1

    microbial_case = base_store.get_cases()[0]
    assert microbial_case.data_analysis == Workflow.MICROSALT
    assert microbial_case.data_delivery == str(DataDelivery.FASTQ_QC)


def test_store_microbial_sample_priority(
    base_store: Store,
    microbial_status_data: dict,
    ticket_id: str,
    store_microbial_order_service: StoreMicrobialOrderService,
):
    # GIVEN a basic store with no samples
    assert not base_store._get_query(table=Sample).first()

    # WHEN storing the order
    store_microbial_order_service.store_items_in_status(
        customer_id=microbial_status_data["customer"],
        order=microbial_status_data["order"],
        ordered=dt.datetime.now(),
        ticket_id=ticket_id,
        items=microbial_status_data["samples"],
        comment="",
        data_analysis=Workflow.MICROSALT,
        data_delivery=DataDelivery.FASTQ_QC,
    )

    # THEN it should store the sample priority
    assert len(base_store._get_query(table=Sample).all()) > 0
    microbial_sample = base_store._get_query(table=Sample).first()

    assert microbial_sample.priority_human == "research"


def test_order_to_status_control_exists(
    sarscov2_order_to_submit: dict,
    base_store: Store,
    store_microbial_order_service: StoreMicrobialOrderService,
):
    # GIVEN sarscov2 order with three samples
    order: OrderIn = OrderIn.parse_obj(sarscov2_order_to_submit, OrderType.SARS_COV_2)

    # WHEN transforming order to status structure
    result: dict = store_microbial_order_service.order_to_status(order=order)

    # THEN check that control is in the result
    sample: dict
    for sample in result.get("samples"):
        assert "control" in sample


def test_order_to_status_control_has_input_value(
    sarscov2_order_to_submit: dict,
    base_store: Store,
    store_microbial_order_service: StoreMicrobialOrderService,
):
    # GIVEN sarscov2 order with three samples with control value set
    control_value = ControlEnum.positive
    order: OrderIn = OrderIn.parse_obj(sarscov2_order_to_submit, OrderType.SARS_COV_2)
    sample: SarsCov2Sample
    for sample in order.samples:
        sample.control: ControlEnum = control_value

    # WHEN transforming order to status structure
    result: dict = store_microbial_order_service.order_to_status(order=order)

    # THEN check that control is in the result
    sample: dict
    for sample in result.get("samples"):
        assert control_value in sample.get("control")


def test_mutant_sample_generates_fields(sarscov2_order_to_submit: dict, base_store: Store):
    """Tests that Mutant orders with region and original_lab set can generate region_code and original_lab_address."""
    # GIVEN sarscov2 order with six samples, one without region_code and original_lab_address

    # WHEN parsing the order
    order: OrderIn = OrderIn.parse_obj(sarscov2_order_to_submit, OrderType.SARS_COV_2)

    # THEN all samples should have region_code and original_lab_address set
    for sample in order.samples:
        assert sample.region_code
        assert sample.original_lab_address


def test_store_items_in_status_control_has_stored_value(
    sarscov2_order_to_submit: dict,
    base_store: Store,
    store_microbial_order_service: StoreMicrobialOrderService,
):
    # GIVEN sarscov2 order with three samples with control value
    order: OrderIn = OrderIn.parse_obj(sarscov2_order_to_submit, OrderType.SARS_COV_2)
    control_value = ControlEnum.positive
    sample: SarsCov2Sample
    for sample in order.samples:
        sample.control: ControlEnum = control_value

    status_data = store_microbial_order_service.order_to_status(order=order)

    # WHEN storing the order
    store_microbial_order_service.store_items_in_status(
        comment="",
        customer_id=order.customer,
        data_analysis=Workflow.MUTANT,
        data_delivery=DataDelivery.FASTQ,
        order="",
        ordered=dt.datetime.now(),
        ticket_id=123456,
        items=status_data.get("samples"),
    )

    # THEN control should exist on the sample in the store
    customer: Customer = base_store.get_customer_by_internal_id(customer_internal_id=order.customer)
    sample: SarsCov2Sample
    for sample in order.samples:
        stored_sample: Sample = base_store.get_sample_by_customer_and_name(
            customer_entry_id=[customer.id], sample_name=sample.name
        )
        assert stored_sample.control == control_value



from pathlib import Path

from cg.io.yaml import read_yaml, read_yaml_stream, write_yaml, write_yaml_stream
from cg.models.mip.mip_sample_info import MipBaseSampleInfo


def test_get_content_from_file(case_qc_sample_info_path: Path):
    """
    Tests read_yaml
    """
    # GIVEN a yaml file

    # WHEN reading the yaml file
    raw_sample_info: dict = read_yaml(file_path=case_qc_sample_info_path)

    # Then assert a dict is returned
    assert isinstance(raw_sample_info, dict)

    # WHEN instantiating a MipBaseSampleInfo object
    sample_info_object = MipBaseSampleInfo(**raw_sample_info)

    # THEN assert that it was successfully created
    assert isinstance(sample_info_object, MipBaseSampleInfo)


def test_get_content_from_stream(yaml_stream: str):
    """
    Tests read_yaml_stream
    """
    # GIVEN a string in yaml format

    # WHEN reading the yaml content in string
    raw_content: list = read_yaml_stream(stream=yaml_stream)

    # THEN assert a list is returned
    assert isinstance(raw_content, list)


def test_write_yaml(case_qc_sample_info_path: Path, cg_dir: Path):
    """
    Tests write_yaml
    """
    # GIVEN a yaml file

    # GIVEN a file path to write to
    yaml_file: Path = Path(cg_dir, "write_yaml.yaml")

    # WHEN reading the yaml file
    raw_sample_info: dict = read_yaml(file_path=case_qc_sample_info_path)

    # WHEN writing the yaml file from dict
    write_yaml(content=raw_sample_info, file_path=yaml_file)

    # THEN assert that a file was successfully created
    assert Path.exists(yaml_file)

    # WHEN reading it as a yaml
    written_raw_sample_info: dict = read_yaml(file_path=yaml_file)

    # THEN assert that all data is kept
    assert raw_sample_info == written_raw_sample_info


def test_write_yaml_stream(yaml_stream: str):
    """
    Tests write_yaml_stream
    """
    # GIVEN a list
    raw_content: list = read_yaml_stream(stream=yaml_stream)

    # WHEN writing the list to a yaml stream
    yaml_content = write_yaml_stream(content=raw_content)

    # THEN assert that all data is kept and properly formatted
    assert yaml_stream == yaml_content



"""Madeline API"""

import logging
import pathlib
import tempfile
from typing import Iterable

from cg.utils import Process

LOG = logging.getLogger(__name__)


class MadelineAPI:
    """Interface to madeline, tool to generate pedigree pictures"""

    def __init__(self, config: dict):
        self.madeline_binary = str(pathlib.Path(config["madeline_exe"]).absolute())
        self.process = Process(binary=self.madeline_binary)

    @staticmethod
    def make_ped(family_id: str, samples: list[dict]) -> Iterable[str]:
        """Yield lines that are used as madeline input."""
        columns = {
            "family": "FamilyId",
            "sample": "IndividualId",
            "sex": "Gender",
            "father": "Father",
            "mother": "Mother",
            "deceased": "Deceased",
            "proband": "Proband",
            "status": "Affected",
        }
        sex_gender = {"male": "M", "female": "F"}
        status_affected = {"affected": "Y", "unaffected": "N"}
        LOG.info("Generating madeline input lines")

        yield "\t".join(columns.values())

        for sample in samples:
            row = [
                family_id,
                sample["sample"],
                sex_gender.get(sample["sex"]) or ".",
                sample.get("father") or ".",
                sample.get("mother") or ".",
                "Y" if sample.get("deceased") else ".",
                "Y" if sample.get("proband") else ".",
                status_affected.get(sample.get("status")) or ".",
            ]
            yield "\t".join(row)

    @staticmethod
    def strip_script_tag(content: str) -> str:
        """Strip away a script tag from a string"""
        script_tag = (
            '<script type="text/javascript" xlink:href=' '"javascript/madeline.js"></script>'
        )
        return content.replace(script_tag, "")

    def run(self, family_id: str, samples: list[dict], out_path: str = None) -> pathlib.Path:
        """Run madeline and generate a file with the results."""
        if out_path:
            out_path = pathlib.Path(out_path)
        else:
            output_dir = pathlib.Path(tempfile.mkdtemp())
            out_path = output_dir / "madeline.xml"

        output_prefix = str(out_path.with_suffix(""))
        LOG.info(f"Generate madeline output to {out_path}")

        ped_stream = self.make_ped(family_id, samples)

        # write the input to a temp file
        with tempfile.NamedTemporaryFile("w") as in_file:
            madeline_content = "\n".join(ped_stream)
            in_file.write(madeline_content)
            in_file.flush()
            madeline_call = [
                "--color",
                "--nolabeltruncation",
                "--outputprefix",
                output_prefix,
                in_file.name,
            ]
            self.process.run_command(parameters=madeline_call)

        with open(out_path, "r") as output:
            svg_content = output.read()

        svg_content = self.strip_script_tag(svg_content)

        with open(out_path, "w") as out_handle:
            out_handle.write(svg_content)

        return out_path

    def __repr__(self):
        return f"MadelineApi({self.madeline_binary})"



"""Priority specific constants"""

from enum import IntEnum, StrEnum


class SlurmQos(StrEnum):
    MAINTENANCE: str = "maintenance"
    LOW: str = "low"
    NORMAL: str = "normal"
    HIGH: str = "high"
    EXPRESS: str = "express"


class TrailblazerPriority(StrEnum):
    LOW: str = "low"
    NORMAL: str = "normal"
    HIGH: str = "high"
    EXPRESS: str = "express"


class PriorityTerms(StrEnum):
    EXPRESS: str = "express"
    PRIORITY: str = "priority"
    RESEARCH: str = "research"
    STANDARD: str = "standard"
    CLINICAL_TRIALS: str = "clinical_trials"


class Priority(IntEnum):
    research: int = 0
    standard: int = 1
    priority: int = 2
    express: int = 3
    clinical_trials: int = 4

    @classmethod
    def priority_to_slurm_qos(cls) -> dict[int, str]:
        return {
            Priority.research: SlurmQos.LOW,
            Priority.standard: SlurmQos.NORMAL,
            Priority.priority: SlurmQos.HIGH,
            Priority.express: SlurmQos.EXPRESS,
            Priority.clinical_trials: SlurmQos.NORMAL,
        }


class SlurmAccount(StrEnum):
    PRODUCTION: str = "production"
    DEVELOPMENT: str = "development"
    STAGE: str = "stage"



"""Delivery report API fixtures."""

import pytest

from cg.meta.delivery_report.raredisease import RarediseaseDeliveryReportAPI
from cg.meta.delivery_report.rnafusion import RnafusionDeliveryReportAPI
from cg.models.cg_config import CGConfig
from cg.models.delivery_report.report import ScoutVariantsFiles


@pytest.fixture
def raredisease_delivery_report_api(
    raredisease_delivery_report_context: CGConfig,
) -> RarediseaseDeliveryReportAPI:
    return RarediseaseDeliveryReportAPI(
        analysis_api=raredisease_delivery_report_context.meta_apis["analysis_api"]
    )


@pytest.fixture
def rnafusion_delivery_report_api(
    rnafusion_delivery_report_context: CGConfig,
) -> RnafusionDeliveryReportAPI:
    return RnafusionDeliveryReportAPI(
        analysis_api=rnafusion_delivery_report_context.meta_apis["analysis_api"]
    )


@pytest.fixture
def scout_variants_files(snv_vcf_file: str, sv_vcf_file: str) -> ScoutVariantsFiles:
    return ScoutVariantsFiles(snv_vcf=snv_vcf_file, sv_vcf=sv_vcf_file)



import re
import shutil

import rich_click as click

from cg.constants import Workflow
from cg.meta.workflow.raredisease import RarediseaseAnalysisAPI
from cg.meta.workflow.rnafusion import RnafusionAnalysisAPI
from cg.meta.workflow.taxprofiler import TaxprofilerAnalysisAPI
from cg.meta.workflow.tomte import TomteAnalysisAPI


def echo_lines(lines: list[str]) -> None:
    for line in lines:
        click.echo(line)


def is_case_name_allowed(name: str) -> bool:
    """Returns true if the given name consists only of letters, numbers and dashes."""
    allowed_pattern: re.Pattern = re.compile("^[A-Za-z0-9-]+$")
    return bool(allowed_pattern.fullmatch(name))


CLICK_CONTEXT_SETTINGS: dict[str, int] = {
    "max_content_width": shutil.get_terminal_size().columns - 10
}


TOWER_WORKFLOW_TO_ANALYSIS_API_MAP: dict = {
    Workflow.RAREDISEASE: RarediseaseAnalysisAPI,
    Workflow.RNAFUSION: RnafusionAnalysisAPI,
    Workflow.TAXPROFILER: TaxprofilerAnalysisAPI,
    Workflow.TOMTE: TomteAnalysisAPI,
}



import pytest

from cg.services.sample_run_metrics_service.sample_run_metrics_service import (
    SampleRunMetricsService,
)
from cg.store.store import Store


@pytest.fixture
def sample_run_metrics_service(
    store_with_illumina_sequencing_data: Store,
) -> SampleRunMetricsService:
    return SampleRunMetricsService(store_with_illumina_sequencing_data)



from pathlib import Path

import pytest

from cg.constants.constants import FileFormat
from cg.io.controller import ReadFile


@pytest.fixture
def balsamic_tga_config_path(fixtures_dir) -> Path:
    """Returns path to a TGA BALSAMIC case_config.json"""

    return Path(fixtures_dir, "apps", "balsamic", "tga_case", "config.json")


@pytest.fixture
def balsamic_tga_metrics_path(fixtures_dir) -> Path:
    """Returns path to a TGA BALSAMIC case_metrics_deliverables.yaml"""

    return Path(fixtures_dir, "apps", "balsamic", "tga_case", "metrics_deliverables.yaml")


@pytest.fixture
def balsamic_tga_config_raw(balsamic_tga_config_path) -> dict:
    """Returna a TGA BALSAMIC config file as a dictionary"""

    config: dict = ReadFile.get_content_from_file(
        file_format=FileFormat.YAML, file_path=balsamic_tga_config_path
    )
    return config


@pytest.fixture
def balsamic_tga_metrics_raw(balsamic_tga_metrics_path) -> dict:
    """Return a TGA BALSAMIC metrics file as a dictionary"""

    metrics: dict = ReadFile.get_content_from_file(
        file_format=FileFormat.YAML, file_path=balsamic_tga_metrics_path
    )
    return metrics


@pytest.fixture
def balsamic_wgs_config_path(fixtures_dir) -> Path:
    """Returns path to a WHOLE_GENOME_SEQUENCING BALSAMIC case_config.json"""

    return Path(fixtures_dir, "apps", "balsamic", "wgs_case", "config.json")


@pytest.fixture
def balsamic_wgs_metrics_path(fixtures_dir) -> Path:
    """Returns path to a WHOLE_GENOME_SEQUENCING BALSAMIC case_metrics_deliverables.yaml"""

    return Path(fixtures_dir, "apps", "balsamic", "wgs_case", "metrics_deliverables.yaml")


@pytest.fixture
def balsamic_wgs_config_raw(balsamic_wgs_config_path) -> dict:
    """Return a WHOLE_GENOME_SEQUENCING BALSAMIC config file as a dictionary"""

    config: dict = ReadFile.get_content_from_file(
        file_format=FileFormat.YAML, file_path=balsamic_wgs_config_path
    )
    return config


@pytest.fixture
def balsamic_wgs_metrics_raw(balsamic_wgs_metrics_path) -> dict:
    """Return a WHOLE_GENOME_SEQUENCING BALSAMIC metrics file as a dictionary"""

    metrics: dict = ReadFile.get_content_from_file(
        file_format=FileFormat.YAML, file_path=balsamic_wgs_metrics_path
    )
    return metrics



"""Tests for running the delivery command"""

import logging
from pathlib import Path

from click.testing import CliRunner

from cg.cli.deliver.base import deliver_case, deliver_ticket
from cg.constants.process import EXIT_SUCCESS, EXIT_FAIL, EXIT_PARSE_ERROR
from cg.models.cg_config import CGConfig
from cg.store.store import Store


def test_run_deliver_with_help(cli_runner: CliRunner, base_context: CGConfig):
    # GIVEN a cli runner and a base context

    # WHEN running the deliver command with help text
    result = cli_runner.invoke(deliver_case, ["--help"], obj=base_context)

    # THEN assert the command exits without problems
    assert result.exit_code == 0


def test_run_deliver_without_specifying_case(cli_runner: CliRunner, base_context: CGConfig, caplog):
    caplog.set_level(logging.INFO)
    # GIVEN a cli runner and a base context
    # WHEN running the deliver command with help text
    result = cli_runner.invoke(deliver_case, obj=base_context)

    # THEN assert the command exits without problems
    assert result.exit_code == EXIT_PARSE_ERROR


def test_run_deliver_non_existing_case(
    cli_runner: CliRunner,
    cg_context: CGConfig,
    case_id: str,
):
    """Test to run the deliver command when the provided case does not exist"""
    # GIVEN a cli runner and a base context

    # GIVEN a case_id that does not exist in the database
    store: Store = cg_context.status_db
    assert store.get_case_by_internal_id(internal_id=case_id) is None

    # WHEN running the deliver command with the non existing case
    result = cli_runner.invoke(
        deliver_case,
        ["--case-id", case_id],
        obj=cg_context,
    )

    # THEN assert the command exits without problems
    assert result.exit_code == 0



"""Module to test fetching of file tags for delivery."""

import pytest

from cg.constants import Workflow
from cg.constants.delivery import PIPELINE_ANALYSIS_TAG_MAP
from cg.services.deliver_files.tag_fetcher.bam_service import (
    BamDeliveryTagsFetcher,
)
from cg.services.deliver_files.tag_fetcher.exc import (
    FetchDeliveryFileTagsError,
)
from cg.services.deliver_files.tag_fetcher.fohm_upload_service import FOHMUploadTagsFetcher
from cg.services.deliver_files.tag_fetcher.models import DeliveryFileTags
from cg.services.deliver_files.tag_fetcher.sample_and_case_service import (
    SampleAndCaseDeliveryTagsFetcher,
)


@pytest.mark.parametrize(
    "workflow",
    [
        Workflow.BALSAMIC,
        Workflow.RAW_DATA,
        Workflow.MIP_DNA,
        Workflow.MIP_RNA,
        Workflow.BALSAMIC_QC,
        Workflow.MICROSALT,
        Workflow.TOMTE,
    ],
)
def test_fetch_tags(workflow: Workflow):
    """Test to fetch tags for the files to deliver."""
    # GIVEN a workflow and a tag fetcher
    tag_fetcher = SampleAndCaseDeliveryTagsFetcher()

    # WHEN fetching the tags for the files to deliver
    tags: DeliveryFileTags = tag_fetcher.fetch_tags(workflow)

    # THEN assert that the tags are fetched
    assert tags.case_tags == PIPELINE_ANALYSIS_TAG_MAP[workflow]["case_tags"]
    assert tags.sample_tags == PIPELINE_ANALYSIS_TAG_MAP[workflow]["sample_tags"]


def test_fetch_tags_unsupported_workflow():
    """Test to fetch tags for the files to deliver."""
    # GIVEN an unsupported workflow and a tag fetcher
    unsupported_workflow: str = "workflow_not_supported"
    tag_fetcher = SampleAndCaseDeliveryTagsFetcher()

    # WHEN fetching the tags for the files to deliver

    # THEN a FetchDeliveryFileTagsError should be raised
    with pytest.raises(FetchDeliveryFileTagsError):
        tag_fetcher.fetch_tags(unsupported_workflow)


def test_bam_delivery_tags_fetcher():
    # GIVEN a workflow and a tag fetcher
    tag_fetcher = BamDeliveryTagsFetcher()

    # WHEN fetching the tags for the files to deliver
    tags: DeliveryFileTags = tag_fetcher.fetch_tags(Workflow.RAW_DATA)

    # THEN assert that the tags are fetched
    assert tags.case_tags is None
    assert tags.sample_tags == [{"bam"}]


def test_fohm_upload_tags_fetcher():
    # GIVEN a tag fetcher
    test_fetcher = FOHMUploadTagsFetcher()

    # WHEN fetching the tags for the files to deliver
    tags: DeliveryFileTags = test_fetcher.fetch_tags(Workflow.MUTANT)

    # THEN assert that the tags are fetched
    assert tags.case_tags is None
    assert tags.sample_tags == [{"consensus-sample"}, {"vcf-report"}]



from datetime import date

from pydantic import BaseModel, conlist
from typing_extensions import Literal


class CrunchyFile(BaseModel):
    path: str
    file: Literal["first_read", "second_read", "spring"]
    checksum: str | None = None
    algorithm: Literal["sha1", "md5", "sha256"] | None = None
    updated: date | None = None


class CrunchyMetadata(BaseModel):
    files: conlist(CrunchyFile, max_length=3, min_length=3)



"""Commands for compressing files, and cleaning uncompressed files"""

import logging

import rich_click as click

from cg.cli.compress.fastq import (
    clean_fastq,
    decompress_case,
    decompress_illumina_run,
    decompress_sample,
    decompress_ticket,
    fastq_cmd,
    fix_spring,
)
from cg.cli.utils import CLICK_CONTEXT_SETTINGS
from cg.meta.backup.backup import SpringBackupAPI
from cg.meta.compress import CompressAPI
from cg.meta.encryption.encryption import SpringEncryptionAPI
from cg.models.cg_config import CGConfig
from cg.services.pdc_service.pdc_service import PdcService

LOG = logging.getLogger(__name__)


@click.group(context_settings=CLICK_CONTEXT_SETTINGS)
@click.pass_obj
def compress(context: CGConfig):
    """Compress files"""

    hk_api = context.housekeeper_api
    crunchy_api = context.crunchy_api

    compress_api = CompressAPI(
        hk_api=hk_api,
        crunchy_api=crunchy_api,
        demux_root=context.run_instruments.illumina.demultiplexed_runs_dir,
    )
    context.meta_apis["compress_api"] = compress_api


compress.add_command(fastq_cmd)


@compress.group()
def clean():
    """Clean uncompressed files"""


clean.add_command(clean_fastq)
clean.add_command(fix_spring)


@click.group(context_settings=CLICK_CONTEXT_SETTINGS)
@click.pass_obj
def decompress(context: CGConfig):
    """Decompress files"""
    hk_api = context.housekeeper_api
    crunchy_api = context.crunchy_api

    pdc_service: PdcService = PdcService(binary_path=context.pdc.binary_path)
    spring_encryption_api: SpringEncryptionAPI = SpringEncryptionAPI(
        binary_path=context.encryption.binary_path,
    )
    spring_backup_api: SpringBackupAPI = SpringBackupAPI(
        encryption_api=spring_encryption_api,
        hk_api=hk_api,
        pdc_service=pdc_service,
    )
    LOG.debug("Start spring retrieval if not dry run")

    compress_api = CompressAPI(
        hk_api=hk_api,
        crunchy_api=crunchy_api,
        demux_root=context.run_instruments.illumina.demultiplexed_runs_dir,
        backup_api=spring_backup_api,
    )

    context.meta_apis["compress_api"] = compress_api
    LOG.info("Running decompress spring")


decompress.add_command(decompress_sample)
decompress.add_command(decompress_case)
decompress.add_command(decompress_ticket)
decompress.add_command(decompress_illumina_run)



from pydantic import BaseModel, Field


class OrderSummary(BaseModel):
    order_id: int = Field(exclude=True)
    total: int
    cancelled: int
    completed: int
    delivered: int
    failed: int
    failed_sequencing_qc: int
    in_lab_preparation: int
    in_sequencing: int
    not_received: int
    running: int



from datetime import datetime

from cg.constants import DataDelivery, Workflow
from cg.models.orders.constants import OrderType
from cg.models.orders.order import OrderIn
from cg.models.orders.sample_base import SexEnum
from cg.services.orders.store_order_services.store_pacbio_order_service import (
    StorePacBioOrderService,
)
from cg.store.models import Case, Sample
from cg.store.store import Store


def test_order_to_status(
    pacbio_order_to_submit: dict, store_pacbio_order_service: StorePacBioOrderService
):
    """Test that a PacBio order is parsed correctly."""
    # GIVEN a PacBio order with two samples
    order = OrderIn.parse_obj(pacbio_order_to_submit, OrderType.PACBIO_LONG_READ)

    # WHEN parsing for status
    data = store_pacbio_order_service.order_to_status(order=order)

    # THEN it should pick out samples and relevant information
    assert len(data["samples"]) == 2
    first_sample = data["samples"][0]
    assert first_sample["name"] == "prov1"
    assert first_sample["application"] == "WGSPCFC060"
    assert first_sample["priority"] == "priority"
    assert first_sample["tumour"] is False
    assert first_sample["volume"] == "25"

    # THEN the other sample is a tumour
    assert data["samples"][1]["tumour"] is True


def test_store_order(
    base_store: Store,
    pacbio_status_data: dict,
    ticket_id: str,
    store_pacbio_order_service: StorePacBioOrderService,
):
    """Test that a PacBio order is stored in the database."""
    # GIVEN a basic store with no samples and a PacBio order
    assert not base_store._get_query(table=Sample).first()
    assert base_store._get_query(table=Case).count() == 0

    # WHEN storing the order
    new_samples: list[Sample] = store_pacbio_order_service._store_samples_in_statusdb(
        customer_id=pacbio_status_data["customer"],
        order=pacbio_status_data["order"],
        ordered=datetime.now(),
        ticket_id=ticket_id,
        samples=pacbio_status_data["samples"],
    )

    # THEN it should store the samples and create a case for each sample
    assert len(new_samples) == 2
    assert len(base_store._get_query(table=Sample).all()) == 2
    assert base_store._get_query(table=Case).count() == 2
    for new_sample in new_samples:
        assert len(new_sample.links) == 1
        case_link = new_sample.links[0]
        assert case_link.case in base_store.get_cases()
        assert case_link.case.data_analysis
        assert case_link.case.data_delivery in [DataDelivery.BAM, DataDelivery.NO_DELIVERY]

    # THEN the sample sex should be stored
    assert new_samples[0].sex == SexEnum.female

    # THEN the analysis for the case should be RAW_DATA
    assert new_samples[0].links[0].case.data_analysis == Workflow.RAW_DATA



from datetime import datetime
from pathlib import Path

from cg.constants import SequencingRunDataAvailability
from cg.constants.demultiplexing import UNDETERMINED
from cg.constants.devices import DeviceType
from cg.models.run_devices.illumina_run_directory_data import IlluminaRunDirectoryData
from cg.services.illumina.data_transfer.models import (
    IlluminaSampleSequencingMetricsDTO,
    IlluminaSequencingRunDTO,
)
from cg.services.illumina.file_parsing.bcl_convert_metrics_parser import (
    BCLConvertMetricsParser,
)
from cg.services.illumina.file_parsing.demux_version_service import (
    IlluminaDemuxVersionService,
)
from cg.services.illumina.file_parsing.sequencing_times.collect_sequencing_times import (
    CollectSequencingTimes,
)


class IlluminaDataTransferService:
    """Service to generate database models related to Illumina flow cell metrics."""

    @staticmethod
    def create_sample_run_metrics_dto(
        sample_internal_id: str,
        lane: int,
        metrics_parser: BCLConvertMetricsParser,
    ) -> IlluminaSampleSequencingMetricsDTO:
        """Create sequencing metrics for all lanes in a flow cell."""
        total_reads: int = metrics_parser.calculate_total_reads_for_sample_in_lane(
            sample_internal_id=sample_internal_id, lane=lane
        )
        q30_bases_percent: float = metrics_parser.get_q30_bases_percent_for_sample_in_lane(
            sample_internal_id=sample_internal_id, lane=lane
        )
        mean_quality_score: float = metrics_parser.get_mean_quality_score_for_sample_in_lane(
            sample_internal_id=sample_internal_id, lane=lane
        )
        yield_: float = metrics_parser.get_yield_for_sample_in_lane(
            sample_internal_id=sample_internal_id, lane=lane
        )
        yield_q30: float = metrics_parser.get_yield_q30_for_sample_in_lane(
            sample_internal_id=sample_internal_id, lane=lane
        )

        return IlluminaSampleSequencingMetricsDTO(
            sample_id=sample_internal_id,
            type=DeviceType.ILLUMINA,
            flow_cell_lane=lane,
            total_reads_in_lane=total_reads,
            base_passing_q30_percent=q30_bases_percent,
            base_mean_quality_score=mean_quality_score,
            yield_=yield_,
            yield_q30=yield_q30,
            created_at=datetime.now(),
        )

    def create_sample_run_dto_for_undetermined_reads(
        self,
        flow_cell: IlluminaRunDirectoryData,
    ) -> list[IlluminaSampleSequencingMetricsDTO]:
        """Return sequencing metrics for any undetermined reads in non-pooled lanes."""
        non_pooled_lanes_and_samples: list[tuple[int, str]] = (
            flow_cell.sample_sheet.get_non_pooled_lanes_and_samples()
        )
        metrics_parser = BCLConvertMetricsParser(flow_cell.path)
        undetermined_metrics: list[IlluminaSampleSequencingMetricsDTO] = []

        for lane, sample_internal_id in non_pooled_lanes_and_samples:
            if not metrics_parser.has_undetermined_reads_in_lane(lane):
                continue

            # Passing Undetermined as the sample id is required to extract the undetermined reads data.
            # BclConvert tags undetermined reads in a lane with the sample id "Undetermined".
            metrics: IlluminaSampleSequencingMetricsDTO = self.create_sample_run_metrics_dto(
                sample_internal_id=UNDETERMINED, lane=lane, metrics_parser=metrics_parser
            )
            metrics.sample_id = sample_internal_id
            undetermined_metrics.append(metrics)
        return undetermined_metrics

    def create_sample_sequencing_metrics_dto_for_flow_cell(
        self,
        flow_cell_directory: Path,
    ) -> list[IlluminaSampleSequencingMetricsDTO]:
        """Parse the demultiplexing metrics data into the sequencing statistics model."""
        metrics_parser = BCLConvertMetricsParser(flow_cell_directory)
        sample_internal_ids: list[str] = metrics_parser.get_sample_internal_ids()
        sample_lane_sequencing_metrics: list[IlluminaSampleSequencingMetricsDTO] = []

        for sample_internal_id in sample_internal_ids:
            for lane in metrics_parser.get_lanes_for_sample(sample_internal_id=sample_internal_id):
                metrics: IlluminaSampleSequencingMetricsDTO = self.create_sample_run_metrics_dto(
                    sample_internal_id=sample_internal_id,
                    lane=lane,
                    metrics_parser=metrics_parser,
                )
                sample_lane_sequencing_metrics.append(metrics)
        return sample_lane_sequencing_metrics

    @staticmethod
    def create_illumina_sequencing_dto(
        demultiplexed_run_dir: IlluminaRunDirectoryData,
    ) -> IlluminaSequencingRunDTO:
        metrics_parser = BCLConvertMetricsParser(demultiplexed_run_dir.path)
        total_reads: int = metrics_parser.get_total_reads_for_flow_cell()
        total_undetermined_reads: int = metrics_parser.get_undetermined_reads_for_flow_cell()
        percent_undetermined_reads: float = (
            metrics_parser.get_percent_undetermined_reads_for_flow_cell()
        )
        percent_q30: float = metrics_parser.get_mean_percent_q30_for_flow_cell()
        mean_quality_score: float = metrics_parser.get_mean_quality_score_sum_for_flow_cell()
        total_yield: int = metrics_parser.get_yield_for_flow_cell()
        yield_q30: int = metrics_parser.get_yield_q30_for_flow_cell()
        cycles: int = demultiplexed_run_dir.run_parameters.get_read_1_cycles()
        software_service = IlluminaDemuxVersionService()
        demux_software: str = software_service.get_demux_software(
            demultiplexed_run_dir.demultiplex_software_info_path
        )
        demux_software_version: str = software_service.get_demux_software_version(
            demultiplexed_run_dir.demultiplex_software_info_path
        )
        collect_sequencing_times = CollectSequencingTimes()
        sequencing_started_at: datetime = collect_sequencing_times.get_start_time(
            run_directory_data=demultiplexed_run_dir
        )
        sequencing_completed_at: datetime = collect_sequencing_times.get_end_time(
            run_directory_data=demultiplexed_run_dir
        )
        demultiplexing_started_at: datetime = demultiplexed_run_dir.demultiplexing_started_at
        demultiplexing_completed_at: datetime = demultiplexed_run_dir.demultiplexing_completed_at
        sequencer_type: str = demultiplexed_run_dir.sequencer_type
        sequencer_name: str = demultiplexed_run_dir.machine_name

        return IlluminaSequencingRunDTO(
            sequencer_type=sequencer_type,
            sequencer_name=sequencer_name,
            data_availability=SequencingRunDataAvailability.ON_DISK,
            archived_at=None,
            has_backup=False,
            total_reads=total_reads,
            total_undetermined_reads=total_undetermined_reads,
            percent_undetermined_reads=percent_undetermined_reads,
            percent_q30=percent_q30,
            mean_quality_score=mean_quality_score,
            total_yield=total_yield,
            yield_q30=yield_q30,
            cycles=cycles,
            demultiplexing_software=demux_software,
            demultiplexing_software_version=demux_software_version,
            sequencing_started_at=sequencing_started_at,
            sequencing_completed_at=sequencing_completed_at,
            demultiplexing_started_at=demultiplexing_started_at,
            demultiplexing_completed_at=demultiplexing_completed_at,
            type=DeviceType.ILLUMINA,
        )



import datetime as dt
import pytest
from cg.exc import OrderError
from cg.models.orders.order import OrderIn, OrderType
from cg.services.orders.store_order_services.store_metagenome_order import (
    StoreMetagenomeOrderService,
)
from cg.store.models import Sample
from cg.store.store import Store


def test_metagenome_to_status(
    metagenome_order_to_submit: dict, store_metagenome_order_service: StoreMetagenomeOrderService
):
    # GIVEN metagenome order with two samples
    order = OrderIn.parse_obj(metagenome_order_to_submit, OrderType.METAGENOME)

    # WHEN parsing for status
    data = store_metagenome_order_service.order_to_status(order=order)
    case = data["families"][0]
    # THEN it should pick out samples and relevant information
    assert len(case["samples"]) == 2
    first_sample = case["samples"][0]
    assert first_sample["name"] == "Bristol"
    assert first_sample["application"] == "METLIFR020"
    assert first_sample["priority"] == "standard"
    assert first_sample["volume"] == "1.0"


def test_store_metagenome_samples_bad_apptag(
    base_store: Store,
    metagenome_status_data: dict,
    ticket_id: str,
    store_metagenome_order_service: StoreMetagenomeOrderService,
):
    # GIVEN a basic store with no samples and a metagenome order
    assert not base_store._get_query(table=Sample).first()

    for sample in metagenome_status_data["families"][0]["samples"]:
        sample["application"] = "nonexistingtag"

    # THEN it should raise OrderError
    with pytest.raises(OrderError):
        # WHEN storing the order
        store_metagenome_order_service.store_items_in_status(
            customer_id=metagenome_status_data["customer"],
            order=metagenome_status_data["order"],
            ordered=dt.datetime.now(),
            ticket_id=ticket_id,
            items=metagenome_status_data["families"],
        )



from cg.constants.gene_panel import GENOME_BUILD_37
from cg.constants.subject import PlinkSex, RelationshipStatus, Sex


def convert_genome_build(value):
    return GENOME_BUILD_37 if value is None else value


def set_parent_if_missing(parent: str | None) -> str:
    return RelationshipStatus.HAS_NO_PARENT if parent is None else parent


def set_sex_if_other(sex: str | None) -> str:
    return PlinkSex.UNKNOWN if sex == Sex.OTHER else sex



from cg.server.dto.orders.orders_request import OrdersRequest
from cg.server.dto.orders.orders_response import Order as OrderResponse, Order
from cg.server.dto.orders.orders_response import OrdersResponse
from cg.services.orders.order_service.models import OrderQueryParams
from cg.services.orders.order_summary_service.dto.order_summary import OrderSummary
from cg.services.orders.order_summary_service.order_summary_service import (
    OrderSummaryService,
)
from cg.store.store import Store
from cg.store.models import Order as DatabaseOrder


class OrderService:
    def __init__(self, store: Store, status_service: OrderSummaryService) -> None:
        self.store = store
        self.summary_service = status_service

    def get_order(self, order_id: int) -> OrderResponse:
        order: Order = self.store.get_order_by_id(order_id)
        summary: OrderSummary = self.summary_service.get_summary(order_id)
        return self._create_order_response(order=order, summary=summary)

    def get_orders(self, orders_request: OrdersRequest) -> OrdersResponse:
        order_query_params: OrderQueryParams = self._get_order_query_params(orders_request)
        orders, total_count = self.store.get_orders(order_query_params)
        order_ids: list[int] = [order.id for order in orders]
        if not order_ids:
            return OrdersResponse(orders=[], total_count=0)
        summaries: list[OrderSummary] = self.summary_service.get_summaries(order_ids)
        return self._create_orders_response(orders=orders, summaries=summaries, total=total_count)

    def set_open(self, order_id: int, open: bool) -> OrderResponse:
        order: Order = self.store.update_order_status(order_id=order_id, open=open)
        return self._create_order_response(order)

    def update_is_open(self, order_id: int, delivered_analyses: int) -> None:
        """Update the is_open parameter of an order based on the number of delivered analyses."""
        order: Order = self.store.get_order_by_id(order_id)
        case_count: int = len(order.cases)
        if self._is_order_closed(case_count=case_count, delivered_analyses=delivered_analyses):
            self.set_open(order_id=order_id, open=False)
        elif not order.is_open:
            self.set_open(order_id=order_id, open=True)

    @staticmethod
    def _get_order_query_params(orders_request: OrdersRequest) -> OrderQueryParams:
        return OrderQueryParams(
            page=orders_request.page,
            page_size=orders_request.page_size,
            search=orders_request.search,
            is_open=orders_request.is_open,
            sort_field=orders_request.sort_field,
            sort_order=orders_request.sort_order,
            workflows=[orders_request.workflow] if orders_request.workflow else [],
        )

    @staticmethod
    def _create_order_response(order: DatabaseOrder, summary: OrderSummary | None = None) -> Order:
        return Order(
            customer_id=order.customer.internal_id,
            ticket_id=order.ticket_id,
            order_date=str(order.order_date.date()),
            id=order.id,
            is_open=order.is_open,
            workflow=order.workflow,
            summary=summary,
        )

    def _create_orders_response(
        self, orders: list[DatabaseOrder], summaries: list[OrderSummary], total: int
    ) -> OrdersResponse:
        orders: list[Order] = [self._create_order_response(order) for order in orders]
        self._add_summaries(orders=orders, summaries=summaries)
        return OrdersResponse(orders=orders, total_count=total)

    @staticmethod
    def _add_summaries(orders: list[Order], summaries: list[OrderSummary]) -> list[Order]:
        order_map = {order.id: order for order in orders}
        for summary in summaries:
            order = order_map[summary.order_id]
            order.summary = summary
        return orders

    @staticmethod
    def _is_order_closed(case_count: int, delivered_analyses: int) -> bool:
        return delivered_analyses >= case_count



""" Test the CLI for mip-rna config-case"""

import logging

from cg.cli.workflow.mip_rna.base import config_case


def test_cg_workflow_mip_rna_config_case_dry_run(cli_runner, caplog, case_id, mip_rna_context):
    """Test print the MIP RNA case config command to console"""

    caplog.set_level(logging.INFO)

    # GIVEN a cli function

    # WHEN we run a case in dry run mode
    result = cli_runner.invoke(config_case, ["--dry-run", case_id], obj=mip_rna_context)

    # THEN the command should be printed
    assert result.exit_code == 0


def test_cg_workflow_mip_rna_config_case(cli_runner, caplog, case_id, mip_rna_context):
    """Test print the MIP rna case config command to file"""

    caplog.set_level(logging.INFO)

    # GIVEN a cli function

    # WHEN we run a case in dry run mode
    result = cli_runner.invoke(config_case, [case_id], obj=mip_rna_context)

    # THEN the command should be printed
    assert result.exit_code == 0
    assert f"Config file saved to" in caplog.text



from pathlib import Path

from cg.services.deliver_files.file_formatter.files.sample_service import LOG
from cg.services.deliver_files.file_formatter.path_name.abstract import PathNameFormatter


class NestedStructurePathFormatter(PathNameFormatter):
    """
    Class to format sample file names and paths in a nested format used to deliver files to a customer inbox.
    """

    def format_file_path(self, file_path: Path, provided_id: str, provided_name: str) -> Path:
        """
        Returns formatted files with original and formatted file names:
        1. Adds a folder with provided name to the path of the files.
        2. Replaces id by name.

        args:
            file_path: The path to the file
            provided_id: The id to replace
            provided_name: The name to replace the id with
        """
        LOG.debug("[FORMAT SERVICE] Formatting sample file names with nested structure.")
        replaced_name = file_path.name.replace(provided_id, provided_name)
        formatted_path = Path(file_path.parent, provided_name, replaced_name)
        return formatted_path



from pathlib import Path

import pytest
from cg.constants.delivery import INBOX_NAME

from cg.constants.constants import Workflow
from cg.models.cg_config import CGConfig
from cg.store.models import Case


@pytest.fixture
def mutant_case(cg_context: CGConfig, case_id: str, ticket_id: str, helpers) -> Case:
    """Return a Mutant case."""
    case = helpers.add_case(
        store=cg_context.status_db,
        internal_id=case_id,
        name=ticket_id,
        data_analysis=Workflow.MUTANT,
    )
    return case


@pytest.fixture
def microsalt_case(cg_context: CGConfig, case_id: str, ticket_id: str, helpers) -> Case:
    """Return a Microsalt case."""
    case = helpers.add_case(
        store=cg_context.status_db,
        internal_id=case_id,
        name=ticket_id,
        data_analysis=Workflow.MICROSALT,
    )
    return case


@pytest.fixture
def destination_path() -> Path:
    """Returns a dummy path."""
    return Path("path", "to", "destination")


@pytest.fixture
def all_samples_in_inbox(
    analysis_family: dict[str, any], dummy_file_name: str, tmpdir_factory, ticket_id: str
) -> Path:
    """Returns a customer inbox path with all samples delivered."""
    inbox = tmpdir_factory.mktemp(INBOX_NAME)
    for index in range(3):
        Path(inbox, ticket_id, analysis_family["samples"][index]["name"]).mkdir(
            exist_ok=True, parents=True
        )
        Path(inbox, ticket_id, analysis_family["samples"][index]["name"], dummy_file_name).touch(
            exist_ok=True
        )
    Path(inbox, ticket_id, analysis_family["name"]).mkdir(exist_ok=True, parents=True)
    Path(inbox, ticket_id, analysis_family["name"], dummy_file_name).touch(exist_ok=True)
    return Path(inbox, ticket_id)


@pytest.fixture
def folders_to_deliver(all_samples_in_inbox) -> set[Path]:
    return set(all_samples_in_inbox.iterdir())


@pytest.fixture
def dummy_file_name() -> str:
    """Returns a dummy file name."""
    return "dummy_file_name"



import subprocess
from pathlib import Path

from cg.apps.slurm.slurm_api import SlurmAPI
from cg.constants import FileExtensions, SequencingRunDataAvailability
from cg.constants.backup import MAX_PROCESSING_ILLUMINA_RUNS
from cg.constants.demultiplexing import DemultiplexingDirsAndFiles
from cg.exc import (
    DsmcAlreadyRunningError,
    IlluminaRunAlreadyBackedUpError,
    IlluminaRunEncryptionError,
    PdcError,
    PdcNoFilesMatchingSearchError,
)
from cg.meta.backup.backup import LOG
from cg.meta.encryption.encryption import EncryptionAPI
from cg.meta.tar.tar import TarAPI
from cg.models.cg_config import PDCArchivingDirectory
from cg.models.run_devices.illumina_run_directory_data import IlluminaRunDirectoryData
from cg.services.illumina.backup.encrypt_service import IlluminaRunEncryptionService
from cg.services.illumina.backup.utils import (
    get_latest_archived_encryption_key_path,
    get_latest_archived_sequencing_run_path,
)
from cg.services.pdc_service.pdc_service import PdcService
from cg.store.models import IlluminaSequencingRun
from cg.store.store import Store
from cg.utils.time import get_elapsed_time, get_start_time


class IlluminaBackupService:
    """Class for retrieving FCs from backup."""

    def __init__(
        self,
        encryption_api: EncryptionAPI,
        pdc_archiving_directory: PDCArchivingDirectory,
        status_db: Store,
        tar_api: TarAPI,
        pdc_service: PdcService,
        sequencing_runs_dir: str,
        dry_run: bool = False,
    ):
        self.encryption_api = encryption_api
        self.pdc_archiving_directory: PDCArchivingDirectory = pdc_archiving_directory
        self.status_db: Store = status_db
        self.tar_api: TarAPI = tar_api
        self.pdc: PdcService = pdc_service
        self.sequencing_runs_dir: str = sequencing_runs_dir
        self.dry_run: bool = dry_run

    def has_processing_queue_capacity(self) -> bool:
        """Check if the processing queue for illumina runs is not full."""
        proccessing_runs_count: int = len(
            self.status_db.get_illumina_sequencing_runs_by_data_availability(
                data_availability=[SequencingRunDataAvailability.PROCESSING]
            )
        )
        LOG.debug(f"Processing illumina runs: {proccessing_runs_count}")
        return proccessing_runs_count < MAX_PROCESSING_ILLUMINA_RUNS

    def get_first_run(self) -> IlluminaSequencingRun | None:
        """Get the sequencing run from the requested queue."""
        sequencing_run: list[IlluminaSequencingRun] | None = (
            self.status_db.get_illumina_sequencing_runs_by_data_availability(
                data_availability=[SequencingRunDataAvailability.REQUESTED]
            )
        )
        return sequencing_run[0] if sequencing_run else None

    def fetch_sequencing_run(
        self, sequencing_run: IlluminaSequencingRun | None = None
    ) -> float | None:
        """Start fetching a sequencing run from backup if possible.
        A run can only be fetched when:
            1. The processing queue is not full.
            2. The requested queue is not emtpy.
        """
        if not self.has_processing_queue_capacity():
            LOG.info("Processing queue is full")
            return None

        if not sequencing_run:
            sequencing_run: IlluminaSequencingRun | None = self.get_first_run()

        if not sequencing_run:
            LOG.info("No sequencing run requested")
            return None

        if not self.dry_run:
            self.status_db.update_illumina_sequencing_run_data_availability(
                sequencing_run=sequencing_run,
                data_availability=SequencingRunDataAvailability.PROCESSING,
            )
            LOG.info(f"{sequencing_run.device.internal_id}: retrieving from PDC")

        dsmc_output: list[str] = self.query_pdc_for_sequencing_run(
            sequencing_run.device.internal_id
        )

        archived_key: Path = get_latest_archived_encryption_key_path(dsmc_output=dsmc_output)
        archived_run: Path = get_latest_archived_sequencing_run_path(dsmc_output=dsmc_output)

        if not self.dry_run:
            return self._process_run(
                sequencing_run=sequencing_run,
                archived_key=archived_key,
                archived_run=archived_run,
            )

    def _process_run(
        self, sequencing_run: IlluminaSequencingRun, archived_key: Path, archived_run: Path
    ) -> float:
        """Process a flow cell from backup. Return elapsed time."""
        start_time: float = get_start_time()
        run_dir = Path(self.sequencing_runs_dir)
        sequencing_run_output_dir = Path(run_dir, archived_run.name.split(".")[0])
        self.retrieve_archived_key(
            archived_key=archived_key, sequencing_run=sequencing_run, run_dir=run_dir
        )
        self.retrieve_archived_sequencing_run(
            archived_run=archived_run, sequencing_run=sequencing_run, run_dir=run_dir
        )

        try:
            (
                decrypted_run,
                encryption_key,
                retrieved_run,
                retrieved_key,
            ) = self.decrypt_sequencing_run(archived_run, archived_key, run_dir)

            self.extract_sequencing_run(decrypted_run, run_dir)
            self.create_rta_complete(sequencing_run_output_dir)
            self.create_copy_complete(sequencing_run_output_dir)
            self.unlink_files(decrypted_run, encryption_key, retrieved_run, retrieved_key)
        except subprocess.CalledProcessError as error:
            LOG.error(f"Decryption failed: {error.stderr}")
            if not self.dry_run:
                self.status_db.update_illumina_sequencing_run_data_availability(
                    sequencing_run=sequencing_run,
                    data_availability=SequencingRunDataAvailability.REQUESTED,
                )
            raise error

        return get_elapsed_time(start_time=start_time)

    def unlink_files(
        self,
        decrypted_run: Path,
        encryption_key: Path,
        retrieved_run: Path,
        retrieved_key: Path,
    ):
        """Remove files after the sequencing run has been fetched from PDC."""
        if self.dry_run:
            return
        LOG.debug("Unlink files")
        message = f"{retrieved_run} not found, skipping removal"
        try:
            retrieved_run.unlink()
        except FileNotFoundError:
            LOG.info(message)
        try:
            decrypted_run.unlink()
        except FileNotFoundError:
            LOG.info(message)
        try:
            retrieved_key.unlink()
        except FileNotFoundError:
            LOG.info(message)
        try:
            encryption_key.unlink()
        except FileNotFoundError:
            LOG.info(message)

    @staticmethod
    def create_rta_complete(flow_cell_directory: Path):
        """Create an RTAComplete.txt file in the flow cell run directory."""
        Path(flow_cell_directory, DemultiplexingDirsAndFiles.RTACOMPLETE).touch()

    @staticmethod
    def create_copy_complete(flow_cell_directory: Path):
        """Create a CopyComplete.txt file in the flow cell run directory."""
        Path(flow_cell_directory, DemultiplexingDirsAndFiles.COPY_COMPLETE).touch()

    def extract_sequencing_run(self, decrypted_run, run_dir):
        """Extract the sequencing run tar archive."""
        extraction_command = self.tar_api.get_extract_file_command(
            input_file=decrypted_run, output_dir=run_dir
        )
        LOG.debug(f"Extract sequencing run command: {extraction_command}")
        self.tar_api.run_tar_command(extraction_command)

    def decrypt_sequencing_run(
        self, archived_run: Path, archived_key: Path, run_dir: Path
    ) -> tuple[Path, Path, Path, Path]:
        """Decrypt the sequencing run."""
        retrieved_key: Path = run_dir / archived_key.name
        encryption_key: Path = retrieved_key.with_suffix(FileExtensions.NO_EXTENSION)
        decryption_command: list[str] = self.encryption_api.get_asymmetric_decryption_command(
            input_file=retrieved_key, output_file=encryption_key
        )
        LOG.debug(f"Decrypt key command: {decryption_command}")
        self.encryption_api.run_gpg_command(decryption_command)
        retrieved_run: Path = run_dir / archived_run.name
        decrypted_run: Path = retrieved_run.with_suffix(FileExtensions.NO_EXTENSION)
        decryption_command: list[str] = self.encryption_api.get_symmetric_decryption_command(
            input_file=retrieved_run,
            output_file=decrypted_run,
            encryption_key=encryption_key,
        )
        LOG.debug(f"Decrypt sequencing run command: {decryption_command}")
        self.encryption_api.run_gpg_command(decryption_command)
        return decrypted_run, encryption_key, retrieved_run, retrieved_key

    def retrieve_archived_key(
        self, archived_key: Path, sequencing_run: IlluminaSequencingRun, run_dir: Path
    ) -> None:
        """Attempt to retrieve an archived key."""
        try:
            self.retrieve_archived_file(
                archived_file=archived_key,
                run_dir=run_dir,
            )
        except PdcError as error:
            LOG.error(f"{sequencing_run.device.internal_id}: key retrieval failed")
            if not self.dry_run:
                self.status_db.update_illumina_sequencing_run_data_availability(
                    sequencing_run=sequencing_run,
                    data_availability=SequencingRunDataAvailability.REQUESTED,
                )
            raise error

    def retrieve_archived_sequencing_run(
        self, archived_run: Path, sequencing_run: IlluminaSequencingRun, run_dir: Path
    ):
        """Attempt to retrieve an archived sequencing run."""
        try:
            self.retrieve_archived_file(
                archived_file=archived_run,
                run_dir=run_dir,
            )
            if not self.dry_run:
                self.status_db.update_illumina_sequencing_run_data_availability(
                    sequencing_run=sequencing_run,
                    data_availability=SequencingRunDataAvailability.RETRIEVED,
                )
        except PdcError as error:
            LOG.error(f"{sequencing_run.device.internal_id}: run directory retrieval failed")
            if not self.dry_run:
                self.status_db.update_illumina_sequencing_run_data_availability(
                    sequencing_run=sequencing_run,
                    data_availability=SequencingRunDataAvailability.REQUESTED,
                )
            raise error

    def query_pdc_for_sequencing_run(self, flow_cell_id: str) -> list[str]:
        """Query PDC for a given flow cell id.
        Raise:
            PdcNoFilesMatchingSearchError if no files are found.
        """
        for _, encryption_directory in self.pdc_archiving_directory:
            search_pattern = f"{encryption_directory}*{flow_cell_id}*{FileExtensions.GPG}"
            self.pdc.query_pdc(search_pattern)
            if self.pdc.was_file_found(self.pdc.process.stderr):
                LOG.info(f"Found archived files for PDC query: {search_pattern}")
                return self.pdc.process.stdout.split("\n")
            LOG.debug(f"No archived files found for PDC query: {search_pattern}")

        raise PdcNoFilesMatchingSearchError(
            message=f"No flow cell files found at PDC for {flow_cell_id}"
        )

    def retrieve_archived_file(self, archived_file: Path, run_dir: Path) -> None:
        """Retrieve the archived file from PDC to a sequencing runs directory."""
        retrieved_file = Path(run_dir, archived_file.name)
        LOG.debug(f"Retrieving file {archived_file} to {retrieved_file}")
        self.pdc.retrieve_file_from_pdc(
            file_path=str(archived_file), target_path=str(retrieved_file)
        )

    def validate_is_run_backup_possible(
        self,
        sequencing_run: IlluminaSequencingRun,
        illumina_run_encryption_service: IlluminaRunEncryptionService,
    ) -> None:
        """Check if back-up of sequencing run is possible.
        Raises:
            DsmcAlreadyRunningError if there is already a Dsmc process ongoing.
            IlluminaRunAlreadyBackupError if sequencing run is already backed up.
            IlluminaRunEncryptionError if encryption is not complete.
        """
        if self.pdc.validate_is_dsmc_running():
            raise DsmcAlreadyRunningError("Too many Dsmc processes are already running")
        if sequencing_run and sequencing_run.has_backup:
            raise IlluminaRunAlreadyBackedUpError(
                f"Sequencing run for flow cell: {sequencing_run.device.internal_id} is already backed-up"
            )
        if not illumina_run_encryption_service.complete_file_path.exists():
            raise IlluminaRunEncryptionError(
                f"Sequencing run for flow cell: {illumina_run_encryption_service.run_dir_data.id} encryption process is not complete"
            )
        LOG.debug("Sequencing run can be backed up")

    def backup_run(
        self, files_to_archive: list[Path], store: Store, sequencing_run: IlluminaSequencingRun
    ) -> None:
        """Back-up sequecing run files."""
        for encrypted_file in files_to_archive:
            if not self.dry_run:
                self.pdc.archive_file_to_pdc(file_path=encrypted_file.as_posix())
        if not self.dry_run:
            store.update_illumina_sequencing_run_has_backup(
                sequencing_run=sequencing_run, has_backup=True
            )
            LOG.info(
                f"Illumina run for flow cell: {sequencing_run.device.internal_id} has been backed up"
            )

    def start_run_backup(
        self,
        sequencing_run: IlluminaSequencingRun,
        run_dir_data: IlluminaRunDirectoryData,
        status_db: Store,
        binary_path: str,
        encryption_dir: Path,
        pigz_binary_path,
        sbatch_parameter,
    ) -> None:
        """Check if back-up of flow cell is possible and if so starts it."""
        illumina_run_encryption_service = IlluminaRunEncryptionService(
            binary_path=binary_path,
            dry_run=self.dry_run,
            encryption_dir=encryption_dir,
            run_dir_data=run_dir_data,
            pigz_binary_path=pigz_binary_path,
            slurm_api=SlurmAPI(),
            sbatch_parameter=sbatch_parameter,
            tar_api=self.tar_api,
        )
        self.validate_is_run_backup_possible(
            sequencing_run=sequencing_run,
            illumina_run_encryption_service=illumina_run_encryption_service,
        )
        self.backup_run(
            files_to_archive=[
                illumina_run_encryption_service.final_passphrase_file_path,
                illumina_run_encryption_service.encrypted_gpg_file_path,
            ],
            store=status_db,
            sequencing_run=sequencing_run,
        )



"""Fixtures related to grouping all sequencing devices: PacBio, Illumina, ONT & Saphire."""

import pytest

from cg.models.cg_config import PostProcessingServices, RunNamesServices
from cg.services.run_devices.pacbio.post_processing_service import PacBioPostProcessingService
from cg.services.run_devices.run_names.pacbio import PacbioRunNamesService


@pytest.fixture
def post_processing_services(
    pac_bio_post_processing_service: PacBioPostProcessingService,
) -> PostProcessingServices:
    """Return the post processing services."""
    return PostProcessingServices(pacbio=pac_bio_post_processing_service)


@pytest.fixture
def run_names_services(pacbio_run_names_service: PacbioRunNamesService) -> RunNamesServices:
    """Return the run names services."""
    return RunNamesServices(pacbio=pacbio_run_names_service)



"""API to prepare fastq files for analysis"""

import logging
from pathlib import Path

from housekeeper.store.models import File, Version

from cg.apps.crunchy import CrunchyAPI
from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.constants import SequencingFileTag
from cg.constants.constants import PIPELINES_USING_PARTIAL_ANALYSES
from cg.meta.compress import files
from cg.meta.compress.compress import CompressAPI
from cg.models import CompressionData
from cg.store.models import Case, Sample
from cg.store.store import Store

LOG = logging.getLogger(__name__)


class PrepareFastqAPI:
    def __init__(self, store: Store, compress_api: CompressAPI):
        self.store: Store = store
        self.hk_api: HousekeeperAPI = compress_api.hk_api
        self.compress_api: CompressAPI = compress_api
        self.crunchy_api: CrunchyAPI = compress_api.crunchy_api

    def get_compression_objects(self, case_id: str) -> list[CompressionData]:
        """Return a list of compression objects"""
        case: Case = self.store.get_case_by_internal_id(internal_id=case_id)
        compression_objects = []
        for link in case.links:
            sample: Sample = link.sample
            sample_id = sample.internal_id
            if self._should_skip_sample(case=case, sample=sample):
                LOG.warning(f"Skipping sample {sample_id} - it has no reads.")
                continue
            version_obj: Version = self.compress_api.hk_api.get_latest_bundle_version(
                bundle_name=sample_id
            )
            compression_objects.extend(files.get_spring_paths(version_obj))
        return compression_objects

    def get_sample_compression_objects(self, sample_id: str) -> list[CompressionData]:
        compression_objects: list[CompressionData] = []
        version_obj: Version = self.compress_api.hk_api.get_latest_bundle_version(
            bundle_name=sample_id
        )
        compression_objects.extend(files.get_spring_paths(version_obj))
        return compression_objects

    def is_sample_decompression_needed(self, sample_id: str) -> bool:
        """Check if decompression is needed for the specified sample."""
        LOG.debug(f"Checking if decompression is needed for {sample_id}.")
        compression_objects = self.get_sample_compression_objects(sample_id=sample_id)
        return any(
            not self.crunchy_api.is_compression_pending(compression_object)
            and not compression_object.pair_exists()
            for compression_object in compression_objects
        )

    @staticmethod
    def _should_skip_sample(case: Case, sample: Sample):
        """
        For some workflows, we want to start a partial analysis disregarding the samples with no reads.
        This method returns true if we should skip the sample.
        """
        if case.data_analysis in PIPELINES_USING_PARTIAL_ANALYSES and not sample.has_reads:
            return True
        return False

    def is_spring_decompression_needed(self, case_id: str) -> bool:
        """Check if spring decompression needs to be started"""
        compression_objects = self.get_compression_objects(case_id=case_id)
        return any(
            not self.crunchy_api.is_compression_pending(compression_object)
            and not compression_object.pair_exists()
            for compression_object in compression_objects
        )

    def is_spring_decompression_running(self, case_id: str) -> bool:
        """Check if case is being decompressed"""
        compression_objects = self.get_compression_objects(case_id=case_id)
        return any(
            self.crunchy_api.is_compression_pending(compression_object)
            for compression_object in compression_objects
        )

    def can_at_least_one_sample_be_decompressed(self, case_id: str) -> bool:
        """Returns True if at least one sample can be decompressed, otherwise False"""
        compression_objects: list[CompressionData] = self.get_compression_objects(case_id=case_id)
        return any(
            self.crunchy_api.is_spring_decompression_possible(compression_object)
            for compression_object in compression_objects
        )

    def add_decompressed_fastq_files_to_housekeeper(self, case_id: str) -> None:
        """Adds decompressed FASTQ files to Housekeeper for a case, if there are any."""
        case: Case = self.store.get_case_by_internal_id(internal_id=case_id)
        for sample in case.samples:
            self.add_decompressed_sample(sample=sample, case=case)

    def add_decompressed_sample(self, sample: Sample, case: Case) -> None:
        """Adds decompressed FASTQ files to Housekeeper for a sample, if there are any."""
        sample_id = sample.internal_id
        if self._should_skip_sample(case=case, sample=sample):
            LOG.warning(f"Skipping sample {sample_id} - it has no reads.")
            return
        version: Version = self.compress_api.hk_api.get_latest_bundle_version(bundle_name=sample_id)
        fastq_files: dict[Path, File] = files.get_hk_files_dict(
            tags=[SequencingFileTag.FASTQ], version_obj=version
        )
        compressions: list[CompressionData] = files.get_spring_paths(version)
        for compression in compressions:
            self.add_decompressed_spring_object(
                compression=compression, fastq_files=fastq_files, sample=sample
            )

    def add_decompressed_spring_object(
        self, compression: CompressionData, fastq_files: dict[Path, File], sample: Case
    ) -> None:
        """Adds decompressed FASTQ files to Housekeeper related to a single spring file."""
        result = True
        sample_id: str = sample.internal_id
        if not self.are_fastqs_in_housekeeper(compression=compression, fastq_files=fastq_files):
            LOG.info(f"Adding FASTQ files to sample {sample_id} in Housekeeper")
            result: bool = self.compress_api.add_decompressed_fastq(sample)
        else:
            LOG.info(
                f"Both {compression.fastq_first} and {compression.fastq_second} "
                f"from sample {sample_id} are already in Housekeeper"
            )
        if not result:
            LOG.warning("Files were not added to fastq!")

    @staticmethod
    def are_fastqs_in_housekeeper(compression: CompressionData, fastq_files: dict[Path, File]):
        return compression.fastq_first in fastq_files and compression.fastq_second in fastq_files



"""Test for the DownSampleMetaData class."""

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.models.downsample.downsample_data import DownsampleData
from cg.store.models import Sample
from cg.store.store import Store


def test_downsample_meta_data_pass_checks(
    store_with_case_and_sample_with_reads: Store,
    downsample_hk_api: HousekeeperAPI,
    downsample_case_internal_id: str,
    downsample_sample_internal_id_1: str,
    downsample_case_name: str,
    number_of_reads_in_millions: int,
    tmp_path_factory,
):
    """Test that the checks pass when initialising a DownsampleData class."""
    # GIVEN a store with a sample and a case
    assert store_with_case_and_sample_with_reads.get_case_by_internal_id(
        internal_id=downsample_case_internal_id
    )
    sample: Sample = store_with_case_and_sample_with_reads.get_sample_by_internal_id(
        internal_id=downsample_sample_internal_id_1
    )

    # GIVEN that the sample has enough reads to down sample
    assert sample.reads > number_of_reads_in_millions * 1_000_000

    # GIVEN a bundle exists in housekeeper for the sample
    assert downsample_hk_api.get_latest_bundle_version(sample.internal_id)

    # WHEN initialising the DownsampleData class
    meta_data = DownsampleData(
        status_db=store_with_case_and_sample_with_reads,
        hk_api=downsample_hk_api,
        sample_id=downsample_sample_internal_id_1,
        case_id=downsample_case_internal_id,
        number_of_reads=number_of_reads_in_millions,
        case_name=downsample_case_name,
        out_dir=tmp_path_factory.mktemp("tmp"),
    )
    # WHEN the number of reads in millionts is converted to a string
    number_of_reads: str = meta_data.convert_number_of_reads_to_string

    # THEN all necessary models to run the down sample command are created
    assert meta_data.downsampled_sample.internal_id == f"{sample.internal_id}DS{number_of_reads}M"
    assert meta_data.downsampled_sample.reads == number_of_reads_in_millions * 1_000_000
    assert meta_data.downsampled_case.name == f"{downsample_case_name}-downsampled"



"""Fixtures for the model tests"""

import gzip
import io
from pathlib import Path

import pytest


@pytest.fixture(name="non_existing_gzipped_file_path")
def non_existing_gzipped_file_path(non_existing_file_path: Path) -> Path:
    """Return the path to a non existing file with gzipped ending"""
    return non_existing_file_path.with_suffix(".gz")


@pytest.fixture(name="filled_gzip_file")
def filled_gzip_file(non_existing_gzipped_file_path: Path, content: str) -> Path:
    """Return the path to a existing file with some content that is gzipped"""
    with gzip.open(non_existing_gzipped_file_path, "wb") as outfile:
        with io.TextIOWrapper(outfile, encoding="utf-8") as enc:
            enc.write(content)
    return non_existing_gzipped_file_path


@pytest.fixture(scope="function")
def gzipped_empty_file(fastq_dir: Path) -> Path:
    """Return a the path to a file that is gzipped without any content

    This file will not have the size = 0 due to gzip header and footer
    """

    return fastq_dir / "dummy_run_R1_001.fastq.gz"



"""This script tests the cli method to set data availability of sequencing runs in StatusDB."""

from _pytest.logging import LogCaptureFixture
from click.testing import CliRunner

from cg.cli.set.base import set_sequencing_run
from cg.constants import EXIT_FAIL, EXIT_SUCCESS, SequencingRunDataAvailability
from cg.models.cg_config import CGConfig
from cg.store.models import IlluminaSequencingRun
from cg.store.store import Store


def test_set_sequencing_run_non_existent_id(
    cli_runner: CliRunner, base_context: CGConfig, caplog: LogCaptureFixture
):
    """Test to set a sequencing run that does not exist."""
    # GIVEN an empty database

    # WHEN trying to set a sequencing run that does not exist
    flow_cell_id: str = "non-existent-id"
    result = cli_runner.invoke(set_sequencing_run, [flow_cell_id], obj=base_context)

    # THEN it should exit successfully but print in console that sequencing run was not found
    assert result.exit_code == EXIT_FAIL
    assert f"Sequencing run with {flow_cell_id} not found" in caplog.text


def test_set_sequencing_run_required(
    cli_runner: CliRunner,
    new_demultiplex_context: CGConfig,
    novaseq_x_flow_cell_id: str,
    caplog: LogCaptureFixture,
):
    """Test to set a sequencing run using only the required arguments."""
    # GIVEN a database with a sequencing run

    # WHEN setting a sequencing run
    result = cli_runner.invoke(
        set_sequencing_run, [novaseq_x_flow_cell_id], obj=new_demultiplex_context
    )

    # THEN it should have been set
    assert result.exit_code == EXIT_FAIL
    assert "Please provide a data availability status" in caplog.text


def test_set_sequencing_run_data_availability(
    cli_runner: CliRunner,
    new_demultiplex_context: CGConfig,
    novaseq_x_flow_cell_id: str,
):
    """Test that the updated sequencing run get the data availability we send in."""
    # GIVEN a database with a sequencing run with a data availability different from 'requested'"
    status_db: Store = new_demultiplex_context.status_db
    data_availability = SequencingRunDataAvailability.statuses()[2]
    run_before_update: IlluminaSequencingRun = (
        status_db.get_illumina_sequencing_run_by_device_internal_id(novaseq_x_flow_cell_id)
    )
    assert run_before_update.data_availability != data_availability

    # WHEN setting the data availability of the sequencing run to 'requested'
    result = cli_runner.invoke(
        set_sequencing_run,
        ["--data-availability", data_availability, novaseq_x_flow_cell_id],
        obj=new_demultiplex_context,
    )

    # THEN the command exits successfully
    assert result.exit_code == EXIT_SUCCESS

    # THEN the data availability of the sequencing run is updated correctly
    updated_run: IlluminaSequencingRun = (
        status_db.get_illumina_sequencing_run_by_device_internal_id(novaseq_x_flow_cell_id)
    )
    assert updated_run.data_availability == data_availability



"""Module for Raredisease analysis API tests."""

import os

from cg.meta.workflow.raredisease import RarediseaseAnalysisAPI
from cg.models.cg_config import CGConfig


def test_get_sample_sheet_content(
    raredisease_context: CGConfig,
    raredisease_case_id: str,
):
    """Test Raredisease nextflow sample sheet creation."""

    # GIVEN Raredisease analysis API
    analysis_api: RarediseaseAnalysisAPI = raredisease_context.meta_apis["analysis_api"]

    # WHEN getting the sample sheet content
    result = analysis_api.get_sample_sheet_content(case_id=raredisease_case_id)

    # THEN return should contain patterns
    patterns = [
        "ADM1",
        "XXXXXXXXX_000000_S000_L001_R1_001.fastq.gz",
        "raredisease_case_enough_reads",
    ]

    contains_pattern = any(
        any(any(pattern in sub_element for pattern in patterns) for sub_element in element)
        for element in result
    )
    assert contains_pattern


def test_write_params_file(raredisease_context: CGConfig, raredisease_case_id: str, mocker):
    # GIVEN Raredisease analysis API and input (nextflow sample sheet path)/output (case directory) parameters
    analysis_api: RarediseaseAnalysisAPI = raredisease_context.meta_apis["analysis_api"]

    # WHEN creating case directory
    analysis_api.create_case_directory(case_id=raredisease_case_id, dry_run=False)

    # THEN care directory is created
    assert os.path.exists(analysis_api.get_case_path(case_id=raredisease_case_id))

    mocker.patch.object(RarediseaseAnalysisAPI, "get_target_bed_from_lims")
    RarediseaseAnalysisAPI.get_target_bed_from_lims.return_value = "some_target_bed_file"

    # WHEN writing parameters file
    analysis_api.write_params_file(case_id=raredisease_case_id)

    # THEN the file is created
    assert os.path.isfile(analysis_api.get_params_file_path(case_id=raredisease_case_id))

    # WHEN writing config file
    analysis_api.create_nextflow_config(case_id=raredisease_case_id)

    # THEN the file is created
    assert os.path.isfile(analysis_api.get_nextflow_config_path(case_id=raredisease_case_id))



import shutil
from pathlib import Path

import pytest
from _pytest.fixtures import FixtureRequest

from cg.constants.pacbio import PacBioDirsAndFiles
from cg.services.run_devices.pacbio.metrics_parser.models import BaseMetrics
from cg.services.run_devices.pacbio.metrics_parser.utils import get_parsed_metrics_from_file_name


@pytest.mark.parametrize(
    "file_name, expected_metrics_fixture",
    [
        (PacBioDirsAndFiles.CONTROL_REPORT, "pac_bio_control_metrics"),
        (PacBioDirsAndFiles.SMRTLINK_DATASETS_REPORT, "pac_bio_smrtlink_databases_metrics"),
        (PacBioDirsAndFiles.CCS_REPORT_SUFFIX, "pac_bio_read_metrics"),
        (PacBioDirsAndFiles.LOADING_REPORT, "pac_bio_productivity_metrics"),
        (PacBioDirsAndFiles.RAW_DATA_REPORT, "pac_bio_polymerase_metrics"),
    ],
    ids=["Control", "Smrtlink-Dataset", "Read", "Productivity", "Polymerase"],
)
def test_get_parsed_metrics_from_file_name(
    file_name: str,
    expected_metrics_fixture: str,
    pacbio_barcoded_report_files_to_parse: list[Path],
    request: FixtureRequest,
):
    """Test the parsing of all PacBio metric files."""
    # GIVEN a correct list of metrics files to parse

    # WHEN parsing the SMRTlink datasets metrics
    parsed_metrics: BaseMetrics = get_parsed_metrics_from_file_name(
        metrics_files=pacbio_barcoded_report_files_to_parse, file_name=file_name
    )

    # THEN the parsed metrics are the expected ones
    expected_metrics: BaseMetrics = request.getfixturevalue(expected_metrics_fixture)
    assert parsed_metrics == expected_metrics


@pytest.mark.parametrize(
    "file_name",
    [
        PacBioDirsAndFiles.CONTROL_REPORT,
        PacBioDirsAndFiles.SMRTLINK_DATASETS_REPORT,
        PacBioDirsAndFiles.CCS_REPORT_SUFFIX,
        PacBioDirsAndFiles.LOADING_REPORT,
        PacBioDirsAndFiles.RAW_DATA_REPORT,
    ],
    ids=["Control", "Smrtlink-Dataset", "Read", "Productivity", "Polymerase"],
)
def test_get_parsed_metrics_from_missing_file(tmp_path: Path, file_name: str):
    """Test that providing a list of non-existing files to the parser raises an error."""
    # GIVEN a list of non-existing paths
    metrics_files: list[Path] = [Path(tmp_path, "non_existing_file.json")]

    # WHEN parsing the metrics
    with pytest.raises(FileNotFoundError):
        # THEN a PacBioMetricsParsingError is raised
        get_parsed_metrics_from_file_name(metrics_files=metrics_files, file_name=file_name)


@pytest.mark.parametrize(
    "metrics_file_name",
    [
        PacBioDirsAndFiles.CONTROL_REPORT,
        PacBioDirsAndFiles.SMRTLINK_DATASETS_REPORT,
        PacBioDirsAndFiles.CCS_REPORT_SUFFIX,
        PacBioDirsAndFiles.LOADING_REPORT,
        PacBioDirsAndFiles.RAW_DATA_REPORT,
    ],
    ids=["Control", "Smrtlink-Dataset", "Read", "Productivity", "Polymerase"],
)
def test_parse_dataset_metrics_validation_error(
    tmp_path: Path,
    metrics_file_name: str,
    pac_bio_wrong_metrics_file: Path,
):
    """Test the parsing of all PacBio metric files with an unexpected error raises an error."""
    # GIVEN a tmp statistics directory with a metrics file without the expected structure
    new_path = Path(tmp_path, metrics_file_name)
    shutil.copyfile(src=pac_bio_wrong_metrics_file, dst=new_path)
    assert new_path.exists()
    metrics_files: list[Path] = [new_path]

    # WHEN parsing the metrics
    with pytest.raises(Exception):
        # THEN a PacBioMetricsParsingError is raised
        get_parsed_metrics_from_file_name(metrics_files=metrics_files, file_name=metrics_file_name)



from pathlib import Path
import pytest

from cg.services.fastq_concatenation_service.fastq_concatenation_service import (
    FastqConcatenationService,
)


@pytest.fixture
def fastq_file_service():
    return FastqConcatenationService()


def create_fastqs_directory(tmp_path: Path):
    fastq_dir = Path(tmp_path, "fastqs")
    fastq_dir.mkdir()
    return fastq_dir


def create_fastq_files(
    fastq_dir: Path, number_forward_reads: int, number_reverse_reads: int, sample_id: str
):
    for i in range(number_forward_reads):
        file = Path(fastq_dir, f"{sample_id}_R1_{i}.fastq.gz")
        file.write_text(f"{sample_id} forward read {i}")

    for i in range(number_reverse_reads):
        file = Path(fastq_dir, f"{sample_id}_R2_{i}.fastq.gz")
        file.write_text(f"{sample_id} reverse read {i}")


@pytest.fixture
def fastqs_dir(tmp_path: Path, sample_id: str) -> Path:
    fastq_dir: Path = create_fastqs_directory(tmp_path=tmp_path)
    create_fastq_files(
        fastq_dir=fastq_dir, number_forward_reads=3, number_reverse_reads=3, sample_id=sample_id
    )
    return fastq_dir


@pytest.fixture
def fastq_dir_existing_concatenated_files(tmp_path: Path, sample_id: str) -> Path:
    fastq_dir: Path = create_fastqs_directory(tmp_path=tmp_path)
    create_fastq_files(
        fastq_dir=fastq_dir, number_forward_reads=3, number_reverse_reads=3, sample_id=sample_id
    )
    forward_output_path = Path(fastq_dir, "forward.fastq.gz")
    reverse_output_path = Path(fastq_dir, "reverse.fastq.gz")
    forward_output_path.write_text("Existing concatenated forward reads")
    reverse_output_path.write_text("Existing concatenated reverse reads")
    return fastq_dir


@pytest.fixture
def fastqs_forward(tmp_path: Path, sample_id: str) -> Path:
    """Return a directory with only forward reads."""
    fastq_dir: Path = create_fastqs_directory(tmp_path=tmp_path)
    create_fastq_files(
        fastq_dir=fastq_dir, number_forward_reads=3, number_reverse_reads=0, sample_id=sample_id
    )
    return fastq_dir


@pytest.fixture
def fastqs_multiple_samples(tmp_path: Path, sample_id: str, another_sample_id: str) -> Path:
    """Return a directory with fastq files for multiple samples."""
    fastq_dir: Path = create_fastqs_directory(tmp_path=tmp_path)
    create_fastq_files(
        fastq_dir=fastq_dir, number_forward_reads=3, number_reverse_reads=3, sample_id=sample_id
    )
    create_fastq_files(
        fastq_dir=fastq_dir,
        number_forward_reads=3,
        number_reverse_reads=3,
        sample_id=another_sample_id,
    )
    return fastq_dir



from pathlib import Path

from cg.services.decompression_service.decompressor import Decompressor


def test_decompressor(zipped_folder: Path, destination_folder: Path, file_in_zip: Path):
    # GIVEN a decompressor, a source path and a destination path
    decompressor = Decompressor()
    assert zipped_folder.exists()

    # WHEN decompressing a file
    decompressed_path = decompressor.decompress(
        source_path=zipped_folder, destination_path=destination_folder
    )

    # THEN assert that the decompressed path exists
    assert decompressed_path.exists()
    assert decompressed_path.is_dir()

    # THEN assert the expected files are in the decompressed path
    assert file_in_zip.name in [file.name for file in decompressed_path.iterdir()]



from cg.constants.priority import SlurmAccount, SlurmQos


def get_quality_of_service(account: SlurmAccount) -> SlurmQos:
    return SlurmQos.HIGH if account == SlurmAccount.PRODUCTION else SlurmQos.LOW



import re
from datetime import datetime
from typing import Any, TypeVar

from pydantic import BaseModel, BeforeValidator, Field, model_validator
from typing_extensions import Annotated

from cg.constants.pacbio import (
    BarcodeMetricsAliases,
    CCSAttributeAliases,
    ControlAttributeAliases,
    LoadingAttributesAliases,
    PolymeraseDataAttributeAliases,
    SampleMetricsAliases,
    SmrtLinkDatabasesAliases,
)
from cg.services.run_devices.abstract_models import RunMetrics
from cg.utils.calculations import fraction_to_percent

BaseMetrics = TypeVar("BaseMetrics", bound=BaseModel)


class ReadMetrics(BaseModel):
    """Model for the read metrics."""

    hifi_reads: int = Field(..., alias=CCSAttributeAliases.HIFI_READS)
    hifi_yield: int = Field(..., alias=CCSAttributeAliases.HIFI_YIELD)
    hifi_mean_read_length: int = Field(..., alias=CCSAttributeAliases.HIFI_MEAN_READ_LENGTH)
    hifi_median_read_length: int = Field(..., alias=CCSAttributeAliases.HIFI_MEDIAN_READ_LENGTH)
    hifi_mean_length_n50: int = Field(..., alias=CCSAttributeAliases.HIFI_READ_LENGTH_N50)
    hifi_median_read_quality: int = Field(..., alias=CCSAttributeAliases.HIFI_MEDIAN_READ_QUALITY)
    percent_q30: float = Field(..., alias=CCSAttributeAliases.PERCENT_Q30)
    failed_reads: int = Field(..., alias=CCSAttributeAliases.FAILED_READS)
    failed_yield: int = Field(..., alias=CCSAttributeAliases.FAILED_YIELD)
    failed_mean_read_length: int = Field(..., alias=CCSAttributeAliases.FAILED_MEAN_READ_LENGTH)


class ControlMetrics(BaseModel):
    """Model for the control metrics."""

    reads: int = Field(..., alias=ControlAttributeAliases.NUMBER_OF_READS)
    mean_read_length: int = Field(..., alias=ControlAttributeAliases.MEAN_READ_LENGTH)
    percent_mean_concordance_reads: Annotated[float, BeforeValidator(fraction_to_percent)] = Field(
        ..., alias=ControlAttributeAliases.PERCENT_MEAN_READ_CONCORDANCE
    )
    percent_mode_concordance_reads: Annotated[float, BeforeValidator(fraction_to_percent)] = Field(
        ..., alias=ControlAttributeAliases.PERCENT_MODE_READ_CONCORDANCE
    )


class ProductivityMetrics(BaseModel):
    """Model for the loading metrics."""

    productive_zmws: int = Field(..., alias=LoadingAttributesAliases.PRODUCTIVE_ZMWS)
    p_0: int = Field(..., alias=LoadingAttributesAliases.P_0)
    p_1: int = Field(..., alias=LoadingAttributesAliases.P_1)
    p_2: int = Field(..., alias=LoadingAttributesAliases.P_2)
    percent_p_0: float
    percent_p_1: float
    percent_p_2: float

    @model_validator(mode="before")
    @classmethod
    def set_percentages(cls, data: Any):
        if isinstance(data, dict):
            productive_zmws = data.get(LoadingAttributesAliases.PRODUCTIVE_ZMWS)
            p_0 = data.get(LoadingAttributesAliases.P_0)
            p_1 = data.get(LoadingAttributesAliases.P_1)
            p_2 = data.get(LoadingAttributesAliases.P_2)
            data["percent_p_0"] = round((p_0 / productive_zmws) * 100, 0)
            data["percent_p_1"] = round((p_1 / productive_zmws) * 100, 0)
            data["percent_p_2"] = round((p_2 / productive_zmws) * 100, 0)
        return data


class PolymeraseMetrics(BaseModel):
    """Model for the polymerase metrics."""

    mean_read_length: int = Field(..., alias=PolymeraseDataAttributeAliases.MEAN_READ_LENGTH)
    read_length_n50: int = Field(..., alias=PolymeraseDataAttributeAliases.READ_LENGTH_N50)
    mean_longest_subread_length: int = Field(
        ..., alias=PolymeraseDataAttributeAliases.MEAN_LONGEST_SUBREAD_LENGTH
    )
    longest_subread_length_n50: int = Field(
        ..., alias=PolymeraseDataAttributeAliases.LONGEST_SUBREAD_LENGTH_N50
    )


class SmrtlinkDatasetsMetrics(BaseModel):
    """Model to parse metrics in the SMRTlink datasets report."""

    cell_id: str = Field(..., alias=SmrtLinkDatabasesAliases.CELL_ID)
    well: str = Field(..., alias=SmrtLinkDatabasesAliases.WELL_NAME)
    well_sample_name: str = Field(..., alias=SmrtLinkDatabasesAliases.WELL_SAMPLE_NAME)
    run_started_at: datetime
    run_completed_at: datetime = Field(..., alias=SmrtLinkDatabasesAliases.RUN_COMPLETED_AT)
    movie_name: str = Field(..., alias=SmrtLinkDatabasesAliases.MOVIE_NAME)
    cell_index: int = Field(..., alias=SmrtLinkDatabasesAliases.CELL_INDEX)
    path: str = Field(..., alias=SmrtLinkDatabasesAliases.PATH)
    plate: int

    @model_validator(mode="before")
    @classmethod
    def extract_plate(cls, data: Any):
        if isinstance(data, dict):
            path = data.get("path")
            if path:
                pattern = r"/([12])_[ABCD]01"
                match = re.search(pattern, path)
                if match:
                    data["plate"] = match.group(1)
        return data

    @model_validator(mode="before")
    @classmethod
    def set_sequencing_completed_at(cls, data: Any):
        if isinstance(data, dict):
            movie_name = data.get(SmrtLinkDatabasesAliases.MOVIE_NAME)
            if movie_name:
                date: str = movie_name.split("_")[1] + movie_name.split("_")[2]
                data["run_started_at"] = datetime.strptime(date, "%y%m%d%H%M%S")
        return data


class BarcodeMetrics(RunMetrics):
    """Model that holds barcode data."""

    barcoded_hifi_reads: int = Field(..., alias=BarcodeMetricsAliases.BARCODED_HIFI_READS)
    barcoded_hifi_reads_percentage: Annotated[float, BeforeValidator(fraction_to_percent)] = Field(
        ..., alias=BarcodeMetricsAliases.BARCODED_HIFI_READS_PERCENTAGE
    )
    barcoded_hifi_yield: int = Field(..., alias=BarcodeMetricsAliases.BARCODED_HIFI_YIELD)
    barcoded_hifi_yield_percentage: Annotated[float, BeforeValidator(fraction_to_percent)] = Field(
        ..., alias=BarcodeMetricsAliases.BARCODED_HIFI_YIELD_PERCENTAGE
    )
    barcoded_hifi_mean_read_length: int = Field(
        ..., alias=BarcodeMetricsAliases.UNBARCODED_HIFI_MEAN_READ_LENGTH
    )
    unbarcoded_hifi_reads: int = Field(..., alias=BarcodeMetricsAliases.UNBARCODED_HIFI_READS)
    unbarcoded_hifi_yield: int = Field(..., alias=BarcodeMetricsAliases.UNBARCODED_HIFI_YIELD)
    unbarcoded_hifi_mean_read_length: int = Field(
        ..., alias=BarcodeMetricsAliases.UNBARCODED_HIFI_MEAN_READ_LENGTH
    )


class SampleMetrics(RunMetrics):
    """Model that holds run data for a specific sample."""

    barcode_name: str = Field(..., alias=SampleMetricsAliases.BARCODE_NAME)
    hifi_mean_read_length: int = Field(..., alias=SampleMetricsAliases.HIFI_MEAN_READ_LENGTH)
    hifi_median_read_quality: str = Field(..., alias=SampleMetricsAliases.HIFI_READ_QUALITY)
    hifi_reads: int = Field(..., alias=SampleMetricsAliases.HIFI_READS)
    hifi_yield: int = Field(..., alias=SampleMetricsAliases.HIFI_YIELD)
    polymerase_read_length: int = Field(..., alias=SampleMetricsAliases.POLYMERASE_READ_LENGTH)
    sample_internal_id: str = Field(..., alias=SampleMetricsAliases.SAMPLE_INTERNAL_ID)


class PacBioMetrics(RunMetrics):
    """Model that holds all relevant PacBio metrics."""

    read: ReadMetrics
    control: ControlMetrics
    productivity: ProductivityMetrics
    polymerase: PolymeraseMetrics
    dataset_metrics: SmrtlinkDatasetsMetrics
    barcodes: BarcodeMetrics
    samples: list[SampleMetrics]



"""Tests for rsync base"""

import logging

from cg.cli.deliver.base import rsync
from cg.constants import EXIT_SUCCESS
from cg.models.cg_config import CGConfig


def test_rsync_help(cli_runner):
    """Test to run the rsync function"""
    # GIVEN a cli runner

    # WHEN running cg deliver rsync
    result = cli_runner.invoke(rsync, ["--dry-run", "--help"])

    # THEN assert the command exists without problems
    assert result.exit_code == EXIT_SUCCESS
    # THEN assert the information is printed
    assert "The folder generated using the" in result.output


def test_run_rsync_command_no_case(cg_context: CGConfig, cli_runner, helpers, caplog):
    """Test generating the rsync command for ticket that doesnt exist"""
    caplog.set_level(logging.INFO)

    # Given an invalid ticket id where case was not created

    # WHEN running deliver rsync command
    result = cli_runner.invoke(rsync, ["--dry-run", "9898989898"], obj=cg_context)

    # THEN command failed successfully
    assert result.exit_code != EXIT_SUCCESS



from datetime import datetime
from pathlib import Path

import pytest

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.constants import SequencingFileTag
from cg.models.cg_config import CGConfig
from cg.services.illumina.cleaning.clean_runs_service import (
    IlluminaCleanRunsService,
)
from cg.models.run_devices.illumina_run_directory_data import IlluminaRunDirectoryData
from cg.store.models import IlluminaSequencingRun
from cg.store.store import Store
from tests.store_helpers import StoreHelpers


@pytest.fixture(scope="function")
def illumina_clean_service_can_be_removed(
    tmp_sequencing_run_to_clean_path: Path,
    store_with_illumina_sequencing_data: Store,
    selected_novaseq_6000_pre_1_5_kits_sample_ids: str,
    housekeeper_api_with_illumina_seq_run_to_clean: HousekeeperAPI,
    tmp_sample_sheet_clean_illumina_sequencing_run_path: Path,
    tmp_sequencing_run_to_clean: IlluminaRunDirectoryData,
) -> IlluminaCleanRunsService:
    """Return a IlluminaCleanRunsService with a sequencing run that can be removed."""
    sequencing_run_to_clean: IlluminaSequencingRun = (
        store_with_illumina_sequencing_data.get_illumina_sequencing_run_by_device_internal_id(
            tmp_sequencing_run_to_clean.id
        )
    )
    sequencing_run_to_clean.has_backup = True

    # Remove a sample metric to fulfill the cleaning criteria
    sample_metric_to_delete = [
        metric
        for metric in sequencing_run_to_clean.sample_metrics
        if metric.sample.internal_id == selected_novaseq_6000_pre_1_5_kits_sample_ids[1]
    ]
    store_with_illumina_sequencing_data.session.delete(sample_metric_to_delete[0])
    clean_flow_cell_api = IlluminaCleanRunsService(
        sequencing_run_path=tmp_sequencing_run_to_clean_path,
        status_db=store_with_illumina_sequencing_data,
        housekeeper_api=housekeeper_api_with_illumina_seq_run_to_clean,
        dry_run=False,
    )
    clean_flow_cell_api.sequencing_run_dir_data._sample_sheet_path_hk = (
        tmp_sample_sheet_clean_illumina_sequencing_run_path
    )
    return clean_flow_cell_api


@pytest.fixture(scope="function")
def illumina_clean_service_can_not_be_removed(
    tmp_sequencing_run_not_to_clean_path: Path,
    store_with_illumina_sequencing_data: Store,
    housekeeper_api_with_sequencing_run_not_to_clean: HousekeeperAPI,
) -> IlluminaCleanRunsService:
    """Return a IlluminaCleanRunsService with a sequencing run that can not be removed."""
    return IlluminaCleanRunsService(
        sequencing_run_path=tmp_sequencing_run_not_to_clean_path,
        status_db=store_with_illumina_sequencing_data,
        housekeeper_api=housekeeper_api_with_sequencing_run_not_to_clean,
        dry_run=False,
    )


@pytest.fixture(scope="function")
def tmp_clean_dir(tmp_path: Path) -> Path:
    """Return a temporary directory for cleaning."""
    clean_path = Path(tmp_path, "clean")
    clean_path.mkdir()
    return clean_path


@pytest.fixture(scope="function")
def tmp_sequencing_run_to_clean_path(
    tmp_novaseq_6000_pre_1_5_kits_flow_cell_path: Path, tmp_clean_dir: Path
) -> Path:
    """Returns the path to a sequencing run fulfilling all cleaning criteria."""
    clean_seq_run_path = Path(tmp_clean_dir, tmp_novaseq_6000_pre_1_5_kits_flow_cell_path.name)
    clean_seq_run_path.mkdir()
    return clean_seq_run_path


@pytest.fixture(scope="function")
def tmp_sequencing_run_to_clean(
    tmp_sequencing_run_to_clean_path: Path,
) -> IlluminaRunDirectoryData:
    """Returns a IlluminaRunDirectoryData object for a sequencing run that fulfills all cleaning criteria."""
    return IlluminaRunDirectoryData(tmp_sequencing_run_to_clean_path)


@pytest.fixture(scope="function")
def tmp_sequencing_run_not_to_clean_path(tmp_novaseq_6000_pre_1_5_kits_flow_cell_path: Path):
    """Return the path to a sequencing run not fulfilling all cleaning criteria."""
    return tmp_novaseq_6000_pre_1_5_kits_flow_cell_path


@pytest.fixture(scope="function")
def tmp_sequencing_run_not_to_clean(
    tmp_sequencing_run_not_to_clean_path: Path,
) -> IlluminaRunDirectoryData:
    """Returns a IlluminaRunDirectoryData object for a sequencing run that does not fulfill all cleaning criteria."""
    return IlluminaRunDirectoryData(tmp_sequencing_run_not_to_clean_path)


@pytest.fixture(scope="session")
def tmp_sample_sheet_clean_illumina_sequencing_run_path(tmp_path_factory) -> Path:
    sample_sheet_path: Path = tmp_path_factory.mktemp("SampleSheet.csv")
    return sample_sheet_path


@pytest.fixture(scope="function")
def housekeeper_api_with_illumina_seq_run_to_clean(
    real_housekeeper_api: HousekeeperAPI,
    helpers: StoreHelpers,
    hk_illumina_sequencing_run_bundle_to_clean: dict,
    hk_sample_bundle_for_illumina_sequencing_run_to_clean: dict,
) -> HousekeeperAPI:
    """
    Return a housekeeper api that contains a flow cell bundle with sample sheet,
    a sample bundle with a fastq and a SPRING file that are tagged with the flow cell.
    """
    helpers.ensure_hk_bundle(
        store=real_housekeeper_api, bundle_data=hk_illumina_sequencing_run_bundle_to_clean
    )
    helpers.ensure_hk_bundle(
        store=real_housekeeper_api,
        bundle_data=hk_sample_bundle_for_illumina_sequencing_run_to_clean,
    )
    return real_housekeeper_api


@pytest.fixture(scope="function")
def housekeeper_api_with_sequencing_run_not_to_clean(
    real_housekeeper_api: HousekeeperAPI,
    helpers: StoreHelpers,
    hk_sample_bundle_for_sequencing_run_not_to_clean: dict,
) -> HousekeeperAPI:
    """
    Return a housekeeper api that contains a flow cell bundle with sample sheet,
    a sample bundle with a fastq and a SPRING file that are tagged with the flow cell name.
    """
    helpers.ensure_hk_bundle(
        store=real_housekeeper_api, bundle_data=hk_sample_bundle_for_sequencing_run_not_to_clean
    )
    return real_housekeeper_api


@pytest.fixture(scope="function")
def hk_illumina_sequencing_run_bundle_to_clean(
    tmp_sequencing_run_to_clean: IlluminaRunDirectoryData,
    timestamp_yesterday: datetime,
    tmp_sample_sheet_clean_illumina_sequencing_run_path: Path,
) -> dict:
    """Housekeeper bundle information for a flow cell that can be cleaned."""
    return {
        "name": tmp_sequencing_run_to_clean.id,
        "created": timestamp_yesterday,
        "expires": timestamp_yesterday,
        "files": [
            {
                "path": str(tmp_sample_sheet_clean_illumina_sequencing_run_path),
                "archive": False,
                "tags": ["samplesheet", tmp_sequencing_run_to_clean.id],
            }
        ],
    }


@pytest.fixture(scope="function")
def hk_sample_bundle_for_illumina_sequencing_run_to_clean(
    selected_novaseq_6000_pre_1_5_kits_sample_ids: str,
    timestamp_yesterday: datetime,
    spring_file: Path,
    fastq_file: Path,
    spring_meta_data_file: Path,
    tmp_sequencing_run_to_clean: IlluminaRunDirectoryData,
) -> dict:
    return {
        "name": selected_novaseq_6000_pre_1_5_kits_sample_ids[0],
        "created": timestamp_yesterday,
        "expires": timestamp_yesterday,
        "files": [
            {
                "path": spring_file.as_posix(),
                "archive": False,
                "tags": [
                    SequencingFileTag.SPRING,
                    selected_novaseq_6000_pre_1_5_kits_sample_ids[0],
                    tmp_sequencing_run_to_clean.id,
                ],
            },
            {
                "path": fastq_file.as_posix(),
                "archive": False,
                "tags": [
                    SequencingFileTag.FASTQ,
                    selected_novaseq_6000_pre_1_5_kits_sample_ids[0],
                    tmp_sequencing_run_to_clean.id,
                ],
            },
            {
                "path": spring_meta_data_file.as_posix(),
                "archive": False,
                "tags": [
                    SequencingFileTag.SPRING_METADATA,
                    selected_novaseq_6000_pre_1_5_kits_sample_ids[0],
                    tmp_sequencing_run_to_clean.id,
                ],
            },
        ],
    }


@pytest.fixture(scope="function")
def hk_sample_bundle_for_sequencing_run_not_to_clean(
    sample_id: str,
    timestamp_yesterday: datetime,
) -> dict:
    return {
        "name": sample_id,
        "created": timestamp_yesterday,
        "expires": timestamp_yesterday,
        "files": [],
    }


@pytest.fixture(scope="function")
def clean_illumina_sequencing_runs_context(
    cg_context: CGConfig,
    tmp_clean_dir: Path,
    tmp_sequencing_run_to_clean: IlluminaRunDirectoryData,
    store_with_illumina_sequencing_data: Store,
    housekeeper_api_with_illumina_seq_run_to_clean: HousekeeperAPI,
    selected_novaseq_6000_pre_1_5_kits_sample_ids: list[str],
) -> CGConfig:
    cg_context.run_instruments.illumina.sequencing_runs_dir = tmp_clean_dir
    cg_context.run_instruments.illumina.demultiplexed_runs_dir = tmp_clean_dir
    cg_context.housekeeper_api_ = housekeeper_api_with_illumina_seq_run_to_clean

    sequencing_run: IlluminaSequencingRun = (
        store_with_illumina_sequencing_data.get_illumina_sequencing_run_by_device_internal_id(
            tmp_sequencing_run_to_clean.id
        )
    )
    sequencing_run.has_backup = True
    sample_metric_to_delete = [
        metric
        for metric in sequencing_run.sample_metrics
        if metric.sample.internal_id == selected_novaseq_6000_pre_1_5_kits_sample_ids[1]
    ]
    store_with_illumina_sequencing_data.session.delete(sample_metric_to_delete[0])
    cg_context.status_db_ = store_with_illumina_sequencing_data

    return cg_context



from sqlalchemy.orm import Query

from cg.store.filters.status_organism_filters import filter_organism_by_internal_id
from cg.store.models import Organism
from cg.store.store import Store


def test_filter_organism_by_internal_id_returns_correct_organism(store_with_organisms: Store):
    """Test filtering an organism by internal ID when the ID exists."""

    # GIVEN a store with multiple organisms
    organisms: Query = store_with_organisms._get_query(table=Organism)
    assert organisms.count() > 0

    # GIVEN a random organism from the store
    organism: Organism = organisms.first()
    assert isinstance(organism, Organism)

    # WHEN filtering the organisms by internal ID
    filtered_organisms: list[Organism] = filter_organism_by_internal_id(
        organisms=organisms, internal_id=organism.internal_id
    ).all()

    # THEN only one organism should be returned
    assert len(filtered_organisms) == 1
    assert isinstance(filtered_organisms[0], Organism)

    # THEN the organism should have the same internal ID as the original organism
    assert filtered_organisms[0].internal_id == organism.internal_id


def test_filter_organism_by_internal_id_returns_empty_list_when_id_does_not_exist(
    store_with_organisms: Store, non_existent_id: str
):
    """Test filtering an organism by internal ID when the ID does not exist."""

    # GIVEN a store with multiple organisms
    organisms: Query = store_with_organisms._get_query(table=Organism)
    assert organisms.count() > 0

    # WHEN filtering the organisms by internal ID
    filtered_organisms: list[Organism] = filter_organism_by_internal_id(
        organisms=organisms, internal_id=non_existent_id
    ).all()

    # THEN the filtered organisms should be an empty list
    assert not filtered_organisms


def test_filter_organism_by_internal_id_returns_empty_list_when_id_is_none(
    store_with_organisms: Store,
):
    """Test filtering an organism by internal ID when the ID is None."""

    # GIVEN a store with multiple organisms
    organisms: Query = store_with_organisms._get_query(table=Organism)
    assert organisms.count() > 0

    # WHEN filtering the organisms by internal ID None
    filtered_organisms: list[Organism] = filter_organism_by_internal_id(
        organisms=organisms, internal_id=None
    ).all()

    # THEN the filtered organisms should be an empty list
    assert not filtered_organisms
